import yahoo_panoptes_snmp as easysnmp
from yahoo_panoptes_snmp.exceptions import *

from .exceptions import *
from .variable import PanoptesSNMPVariable
from ...plugins.context import PanoptesPluginContext
from ...validators import PanoptesValidators

exception_class_mapping = {EasySNMPError: PanoptesSNMPException,
                           EasySNMPConnectionError: PanoptesSNMPConnectionException,
                           EasySNMPTimeoutError: PanoptesSNMPTimeoutException,
                           EasySNMPUnknownObjectIDError: PanoptesSNMPUnknownObjectIDException,
                           EasySNMPNoSuchNameError: PanoptesSNMPNoSuchNameException,
                           EasySNMPNoSuchObjectError: PanoptesSNMPNoSuchObjectException,
                           EasySNMPNoSuchInstanceError: PanoptesSNMPNoSuchInstanceException,
                           EasySNMPUndeterminedTypeError: PanoptesSNMPUndeterminedTypeException}


SNMP_ERRORS_MAP = {
    'ERROR': PanoptesSNMPException,
    'CONNECTION': PanoptesSNMPConnectionException,
    'TIMEOUT': PanoptesSNMPTimeoutException,
    'NOSUCHOBJECT': PanoptesSNMPNoSuchObjectException,
    'NOSUCHINSTANCE': PanoptesSNMPNoSuchInstanceException,
    'NOSUCHNAME': PanoptesSNMPNoSuchNameException,
    'UNKNOWNOBJECTID': PanoptesSNMPUnknownObjectIDException,
    'UNDETERMINEDTYPE': PanoptesSNMPUndeterminedTypeException
}


class PanoptesSNMPConnection(object):
    def __init__(self, host, port, timeout, retries):
        assert PanoptesValidators.valid_nonempty_string(host), 'host must a non-empty string'
        assert PanoptesValidators.valid_port(port), 'port must be an integer between 1 and 65535'
        assert PanoptesValidators.valid_nonzero_integer(timeout), 'timeout must be a integer greater than zero'
        assert PanoptesValidators.valid_positive_integer(retries), 'retries must a non-negative integer'

        self._host = host
        self._port = port
        self._timeout = timeout
        self._retries = retries
        self._easy_snmp_session = None

    @staticmethod
    def _wrap_panoptes_snmp_exception(easy_snmp_exception):
        """
        Maps EasySNMP exceptions to PanoptesSNMPExceptions

        Args:
            easy_snmp_exception (Exception): The EasySNMP exception to map

        Returns:
            PanoptesSNMPException: An equivalent PanoptesSNMPException
        """
        try:
            return exception_class_mapping[type(easy_snmp_exception)](easy_snmp_exception.message)
        except KeyError:
            return PanoptesSNMPException(repr(easy_snmp_exception))

    def get(self, oid):
        """
        Takes a single numeric oid and returns a PanoptesSNMPVariable containing the queried oid, index, value and type

        Args:
            oid (str): A numeric oid

        Returns:
            PanoptesSNMPVariable: A PanoptesSNMPVariable containing the queried oid, index, value and type
        """
        assert PanoptesValidators.valid_numeric_snmp_oid(oid), 'oid must be numeric string with a leading period'
        assert self._easy_snmp_session is not None, 'PanoptesSNMPSession not initialized'

        try:
            varbind = self._easy_snmp_session.get(oids=oid)
            return PanoptesSNMPVariable(queried_oid=oid, oid=varbind.oid, index=varbind.oid_index, value=varbind.value,
                                        snmp_type=varbind.snmp_type)
        except Exception as e:
            raise self._wrap_panoptes_snmp_exception(e)

    def get_bulk(self, oid, non_repeaters=0, max_repetitions=10, base_oid=None):
        """
         Takes a single numeric oid and returns a list of PanoptesSNMPVariables which start with the oid

        Args:
            oid (str): A numeric oid
            non_repeaters (int): A non-negative integer
            max_repetitions (int):  A non-negative integer
            base_oid (str): An optional numeric oid which is used to set the oid in the returned PanoptesSNMPVariables

        Returns:
            list(): A list of PanoptesSNMPVariables starting with the provided oid
        """
        assert PanoptesValidators.valid_numeric_snmp_oid(oid), 'oid must be numeric string with a leading period'
        assert PanoptesValidators.valid_positive_integer(non_repeaters), 'non_repeaters must a positive integer'
        assert PanoptesValidators.valid_positive_integer(max_repetitions), 'max_repetitions must a positive integer'
        assert (base_oid is None) or PanoptesValidators.valid_numeric_snmp_oid(
                base_oid), 'oid must be numeric string with a leading period'
        assert self._easy_snmp_session is not None, 'PanoptesSNMPSession not initialized'

        try:
            varbinds = self._easy_snmp_session.get_bulk(oids=oid, non_repeaters=non_repeaters,
                                                        max_repetitions=max_repetitions)

            result = list()
            for varbind in varbinds:
                if varbind.oid.startswith(base_oid):
                    result.append(PanoptesSNMPVariable(queried_oid=base_oid if base_oid else oid, oid=varbind.oid,
                                                       index=varbind.oid_index, value=varbind.value,
                                                       snmp_type=varbind.snmp_type))
            return result
        except Exception as e:
            raise self._wrap_panoptes_snmp_exception(e)

    def bulk_walk(self, oid, non_repeaters=0, max_repetitions=10):
        """
        Takes a single numeric oid and fetches oids under it's subtree using GETBULKs

        Args:
            oid (str): A numeric oid
            non_repeaters (int): A non-negative integer
            max_repetitions (int):  A non-negative integer

        Returns:
            list(PanoptesSNMPVariable): A list of PanoptesSNMPVariables within the subtree of the provided oid
        """
        assert PanoptesValidators.valid_numeric_snmp_oid(oid), 'oid must be numeric string with a leading period'
        assert PanoptesValidators.valid_positive_integer(non_repeaters), 'non_repeaters must a positive integer'
        assert PanoptesValidators.valid_positive_integer(max_repetitions), 'max_repetitions must a positive integer'
        assert self._easy_snmp_session is not None, 'PanoptesSNMPSession not initialized'

        base_oid = oid
        varbinds = list()
        running = 1

        while running:
            results = self.get_bulk(oid=oid, non_repeaters=non_repeaters,
                                    max_repetitions=max_repetitions, base_oid=base_oid)
            varbinds.extend(results)
            if len(results) < max_repetitions:
                break
            oid = base_oid + '.' + results[-1].index

        return varbinds


class PanoptesSNMPV2Connection(PanoptesSNMPConnection):
    def __init__(self, host, port, timeout, retries, community):
        """
        Starts a SNMP V2 connection with the given parameters

        Args:
            host (str): The host to interact with SNMP
            port (int): The port on the host
            timeout (int): Non-zero seconds to wait for connection and replies
            retries (int): The number of times to retry a failed query
            community (str): The community string to use

        Returns:
            None: The session is initiated and stored locally
        """

        super(PanoptesSNMPV2Connection, self).__init__(host, port, timeout, retries)
        assert PanoptesValidators.valid_nonempty_string(community), 'community_string must a non-empty string'

        self._community = community

        try:
            self._easy_snmp_session = easysnmp.Session(hostname=self._host, remote_port=self._port,
                                                       timeout=self._timeout, retries=self._retries,
                                                       version=2, use_numeric=True,
                                                       community=self._community)

        except Exception as e:
            raise self._wrap_panoptes_snmp_exception(e)


class PanoptesSNMPV3TLSConnection(PanoptesSNMPConnection):
    def __init__(self, host, port, timeout, retries, context, our_identity, their_identity):
        """
        Starts a SNMP V3 connection with the given parameters - works only with TLS

        Args:
            host (str): The host to interact with SNMP
            port (int): The port on the host
            timeout (int): Non-zero seconds to wait for connection and replies
            retries (int): The number of times to retry a failed query
            context (str): A non-empty SNMP v3 context
            our_identity (str): The fingerprint or filename of the certificate to present to the host
            their_identity (str): The fingerprint or filename of the certificate to expect from the host

        Returns:
            None: The session is initiated and stored locally
        """
        super(PanoptesSNMPV3TLSConnection, self).__init__(host, port, timeout, retries)
        assert PanoptesValidators.valid_nonempty_string(context), 'context must be a non-empty string'
        assert PanoptesValidators.valid_nonempty_string(our_identity), 'our_identity must be a non-empty string'
        assert PanoptesValidators.valid_nonempty_string(their_identity), 'their_identity must be a non-empty string'

        self._context = context
        self._our_identity = our_identity
        self._their_identity = their_identity

        try:
            self._easy_snmp_session = easysnmp.Session(hostname=self._host, remote_port=self._port,
                                                       timeout=self._timeout, retries=self._retries,
                                                       version=3, use_numeric=True,
                                                       transport='tlstcp',
                                                       context=self._context,
                                                       our_identity=self._our_identity,
                                                       their_identity=self._their_identity)
        except Exception as e:
            raise self._wrap_panoptes_snmp_exception(e)


class PanoptesSNMPPluginConfiguration(object):
    """
    This class encapsulates all SNMP related configuration, including fetching SNMP community string from the secrets
    store as needed

    Args:
        plugin_context: The PanoptesPluginContext from which to derive all SNMP configuration
    """
    def __init__(self, plugin_context):
        assert isinstance(plugin_context,
                          PanoptesPluginContext), 'plugin_context must be a class or subclass of ' \
                                                  'PanoptesPluginContext'

        self._plugin_context = plugin_context
        self._plugin_snmp_configuration = plugin_context.config.get('snmp', {})
        self._default_snmp_configuration = plugin_context.snmp

        self._connection_factory_module = None
        self._connection_factory_class = None
        self._community_string_key = None
        self._proxy_port = None
        self._port = None
        self._community = None
        self._community_string_key = None
        self._timeout = None
        self._retries = None
        self._non_repeaters = None
        self._max_repetitions = None

        self._parse_snmp_configuration()
        self._get_snmp_community_string()

    def _get_snmp_community_string(self):
        """
        This method sets the community string that should be used. It looks up the community string in the following
        order of preference

        1. A community string specified in the plugin configuration file
        2. A community string related to the site in which the resource resides from the 'secrets' store
        3. The default community string specified in the Panoptes configuration file

        Returns:
            None
        """
        logger = self._plugin_context.logger

        # Try and get community string from the plugin configuration
        self._community = self._plugin_snmp_configuration.get('community')

        if self._community:
            assert PanoptesValidators.valid_nonempty_string(
                    self._community), 'SNMP community must be a non-empty string'
            return

        # Else lookup the site-specific community string
        site = self._plugin_context.data.resource_site

        try:
            logger.debug('Going to get fetch SNMP community string using key "{}" for site "{}"'.format(
                    self.community_string_key, site))
            self._community = self._plugin_context.secrets.get_by_site(self.community_string_key, site)
        except Exception as e:
            raise PanoptesSNMPException(
                    'Could not fetch SNMP community string for site "{}" using key "{}": {}'.format(
                            site, self.community_string_key, repr(e)))

        if self._community:
            assert PanoptesValidators.valid_nonempty_string(
                    self._community), 'SNMP community must be a non-empty string'
            return

        # Else return default community string
        self._community = self._default_snmp_configuration.get('community')

        assert PanoptesValidators.valid_nonempty_string(
                self._community), 'SNMP community must be a non-empty string'

    def _parse_snmp_configuration(self):
        self._connection_factory_module = self._plugin_snmp_configuration.get('connection_factory_module',
                                                                              self._default_snmp_configuration[
                                                                                  'connection_factory_module'])
        assert PanoptesValidators.valid_nonempty_string(
                self._connection_factory_module), 'SNMP connection factory module must be a non-empty string'

        self._connection_factory_class = self._plugin_snmp_configuration.get('connection_factory_class',
                                                                             self._default_snmp_configuration[
                                                                                 'connection_factory_class'])
        assert PanoptesValidators.valid_nonempty_string(
                self._connection_factory_class), 'SNMP connection factory class must be a non-empty string'

        self._community_string_key = self._plugin_snmp_configuration.get('community_string_key',
                                                                         self._default_snmp_configuration.get(
                                                                                 'community_string_key'))
        assert PanoptesValidators.valid_nonempty_string(
                self._community_string_key), 'SNMP community string key must be a non-empty string'

        self._port = int(
                self._plugin_snmp_configuration.get('port', self._default_snmp_configuration['port']))
        assert PanoptesValidators.valid_port(self._port), 'SNMP port must be a valid TCP/UDP port number'

        self._proxy_port = int(
                self._plugin_snmp_configuration.get('proxy_port', self._default_snmp_configuration['proxy_port']))
        assert PanoptesValidators.valid_port(self._proxy_port), 'SNMP proxy port must be a valid TCP/UDP port number'

        self._timeout = int(
                self._plugin_snmp_configuration.get('timeout', self._default_snmp_configuration['timeout']))
        assert PanoptesValidators.valid_nonzero_integer(
                self._timeout), 'SNMP timeout must be a positive integer'

        self._retries = int(
                self._plugin_snmp_configuration.get('retries', self._default_snmp_configuration['retries']))
        assert PanoptesValidators.valid_nonzero_integer(
                self._retries), 'SNMP retries must be a positive integer'

        self._non_repeaters = int(self._plugin_snmp_configuration.get('non_repeaters',
                                                                      self._default_snmp_configuration[
                                                                          'non_repeaters']))
        assert PanoptesValidators.valid_positive_integer(
                self._non_repeaters), 'SNMP non-repeaters must be a positive integer'

        self._max_repetitions = int(
                self._plugin_snmp_configuration.get('max_repetitions',
                                                    self._default_snmp_configuration['max_repetitions']))
        assert PanoptesValidators.valid_nonzero_integer(
                self._max_repetitions), 'SNMP max-repetitions must be a positive integer'

    @property
    def connection_factory_module(self):
        return self._connection_factory_module

    @property
    def connection_factory_class(self):
        return self._connection_factory_class

    @property
    def community_string_key(self):
        return self._community_string_key

    @property
    def community(self):
        return self._community

    @property
    def port(self):
        return self._port

    @property
    def proxy_port(self):
        return self._proxy_port

    @property
    def timeout(self):
        return self._timeout

    @property
    def retries(self):
        return self._retries

    @property
    def non_repeaters(self):
        return self._non_repeaters

    @property
    def max_repetitions(self):
        return self._max_repetitions


class PanoptesSNMPConnectionFactory(object):
    """
    This class returns an SNMP connection object. Currently only supports V2 connection objects
    """
    @staticmethod
    def get_snmp_connection(plugin_context, snmp_configuration, community_suffix=None):
        assert isinstance(plugin_context,
                          PanoptesPluginContext), 'plugin_context must be a class or subclass of ' \
                                                  'PanoptesPluginContext'
        assert isinstance(snmp_configuration,
                          PanoptesSNMPPluginConfiguration), 'plugin_context must be a class or subclass of ' \
                                                            'PanoptesSNMPPluginConfiguration'

        host = plugin_context.data.resource_endpoint
        community = snmp_configuration.community

        try:
            if community_suffix:
                community += '@' + str(community_suffix)

            return PanoptesSNMPV2Connection(host=host, port=snmp_configuration.port,
                                            timeout=snmp_configuration.timeout,
                                            retries=snmp_configuration.retries,
                                            community=community)
        except Exception as e:
            raise PanoptesSNMPConnectionException(
                    'Error while creating SNMP connection for device "{}": {}'.format(host, repr(e)))



"""
Copyright 2018, Oath Inc.
Licensed under the terms of the Apache 2.0 license. See LICENSE file in project root for terms.
"""

import unittest

import yahoo_panoptes.polling.polling_plugin_agent


class TestPollingPluginAgent(unittest.TestCase):
    def test_polling_plugin_task(self):
        self.assertTrue(True)



"""
Base mib class
"""
from ....metrics import PanoptesMetricType


class SNMPTypeMixin(object):
    name = 'float'
    metric_type = PanoptesMetricType.GAUGE
    valid_types = float

    def __init__(self, value):
        self.value = value
        self.validate()

    def __str__(self):
        return self.name

    def validate(self):
        """
        Validate that the value is valid

        Raises
        ------
        ValueError - The value is invalid
        """
        self.validate_type()
        self.validate_value()

    def validate_type(self):
        """
        Verify the value is the right type

        Raises
        ------
        ValueError - The self.value is not the right type
        """
        if not isinstance(self.value, self.valid_types):
            raise ValueError('%r is not a %s' % type(self.valid_types))

    def validate_value(self):
        """
        Validate the value of value is valid

        Raises
        ------
        ValueError - The self.value is not valid
        """
        pass


class SNMPType(SNMPTypeMixin):
    pass


class SNMPString(str, SNMPTypeMixin):
    pass


class SNMPInteger(int, SNMPTypeMixin):
    valid_types = int
    metric_type = PanoptesMetricType.COUNTER


class SNMPInteger32(SNMPInteger):
    def validate_value(self):
        if self.value > 65535:
            raise ValueError('32 bit integer overflow, value of %r is to large for a 32 bit integer' % self.value)


class SNMPGauge32(SNMPInteger32):
    pass


class SNMPFloat(float, SNMPTypeMixin):
    pass


class oid(object):
    """
    An mib oid class
    """
    def __init__(self, value, snmp_type=None, description=None):
        """
        Parameters
        ----------
        value: str
            Initial oid string value
        """
        self.value = value
        self.snmp_type = snmp_type
        self.description = description

    @property
    def oid(self):
        return self.value

    def __repr__(self):
        return 'oid("' + str(self.value) + '")'

    def __str__(self):
        return self.value

    def __add__(self, item):
        try:
            snmp_type = item.snmp_type
        except AttributeError:
            snmp_type = self.snmp_type
        item = str(item).strip('.')
        return oid(str(self.value) + '.' + item, snmp_type=snmp_type)


class Mib(object):
    """
    Base Mib class
    """
    pass



"""
Copyright 2018, Oath Inc.
Licensed under the terms of the Apache 2.0 license. See LICENSE file in project root for terms.
"""

import unittest
import json
import os

from mock import patch, Mock
from tests.test_framework import PanoptesMockRedis

from yahoo_panoptes.framework.context import PanoptesContext
from yahoo_panoptes.framework.resources import PanoptesResource
from yahoo_panoptes.framework.enrichment import PanoptesEnrichmentCacheError, PanoptesEnrichmentCache, \
    PanoptesEnrichmentCacheKeyValueStore

mock_time = Mock()
mock_time.return_value = 1521668419.881953


def ordered(obj):
    if isinstance(obj, dict):
        return sorted((k, ordered(v)) for k, v in obj.items())
    if isinstance(obj, list):
        return sorted(ordered(x) for x in obj)
    else:
        return obj


def _get_test_conf_file():
    my_dir = os.path.dirname(os.path.realpath(__file__))
    panoptes_test_conf_file = os.path.join(my_dir, 'config_files/test_panoptes_config.ini')

    return my_dir, panoptes_test_conf_file


class MockLogger(object):
    def __init__(self):
        self._mock_error = Mock()
        self._mock_debug = Mock()

    @property
    def error(self):
        return self._mock_error

    @property
    def debug(self):
        return self._mock_debug


class TestPanoptesEnrichmentCache(unittest.TestCase):
    @patch('redis.StrictRedis', PanoptesMockRedis)
    def setUp(self):
        self.my_dir, self.panoptes_test_conf_file = _get_test_conf_file()
        self._panoptes_context = PanoptesContext(self.panoptes_test_conf_file,
                                                 key_value_store_class_list=[PanoptesEnrichmentCacheKeyValueStore])

        self._enrichment_kv = self._panoptes_context.get_kv_store(PanoptesEnrichmentCacheKeyValueStore)

        self._panoptes_resource = PanoptesResource(resource_site='test_site',
                                                   resource_class='test_class',
                                                   resource_subclass='test_subclass',
                                                   resource_type='test_type',
                                                   resource_id='test_resource_id',
                                                   resource_endpoint='test_endpoint',
                                                   resource_plugin='test_plugin')

        self._plugin_conf = {'Core': {'name': 'Heartbeat Enrichment Plugin',
                                      'module': 'plugin_enrichment_heartbeat'},
                             'main':
                                 {'execute_frequency': '60',
                                  'enrichment_ttl': '300',
                                  'resource_filter': 'resource_class = "system" AND resource_subclass = '
                                                     '"internal" AND resource_type = "heartbeat"'},
                             'enrichment':
                                 {'preload': 'self:test_namespace'}
                             }

        self._enrichment_kv.set('test_resource_id:test_namespace',
                                '{{"data": {{"heartbeat": {{"timestamp": 1521668419}}}}, '
                                '"metadata": {{"_enrichment_group_creation_timestamp": {:.5f}, '
                                '"_enrichment_ttl": 300, "_execute_frequency": 60}}}}'.format(mock_time.return_value))
        self._enrichment_kv.set(
            'test_resource_id01:interface',
            '{{"data": {{"1226": {{"ifType": "l3ipvlan", "ifAlias": "<not set>", "ifPhysAddress": '
            '"00:1f:a0:13:8c:16", "ifDescr": "Virtual Ethernet 226", "ifName": "Virtual Ethernet 226", '
            '"ifMtu": "1500"}}, "1209": {{"ifType": "l3ipvlan", "ifAlias": "<not set>", "ifPhysAddress": '
            '"00:1f:a0:13:8c:15", "ifDescr": "Virtual Ethernet 209", "ifName": "Virtual Ethernet 209", '
            '"ifMtu": "1500"}}}}, "metadata": {{"_enrichment_group_creation_timestamp": {:.5f}, '
            '"_enrichment_ttl": 300, "_execute_frequency": 60}}}}'.format(mock_time.return_value))

        self._enrichment_kv.set(
            'test_resource_id02:interface',
            '{{"data": {{"2226": {{"ifType": "l3ipvlan", "ifAlias": "<not set>", "ifPhysAddress": '
            '"00:1f:a0:13:8c:16", "ifDescr": "Virtual Ethernet 326", "ifName": "Virtual Ethernet 326", '
            '"ifMtu": "1500"}}, "2209": {{"ifType": "l3ipvlan", "ifAlias": "<not set>", "ifPhysAddress": '
            '"00:1f:a0:13:8c:15", "ifDescr": "Virtual Ethernet 309", "ifName": "Virtual Ethernet 309", '
            '"ifMtu": "1500"}}}}, "metadata": {{"_enrichment_group_creation_timestamp": {:.5f}, '
            '"_enrichment_ttl": 300, "_execute_frequency": 60}}}}'.format(mock_time.return_value))

        self._enrichment_kv.set('asn_api:namespace01',
                                '{{"data": {{"asn_data01": {{"data": 101}}}}, '
                                '"metadata": {{"_enrichment_group_creation_timestamp": {:.5f}, '
                                '"_enrichment_ttl": 300, "_execute_frequency": 60}}}}'.format(mock_time.return_value))
        self._enrichment_kv.set('asn_api:namespace02',
                                '{{"data": {{"asn_data02": {{"data": 202}}}}, '
                                '"metadata": {{"_enrichment_group_creation_timestamp": {:.5f}, '
                                '"_enrichment_ttl": 300, "_execute_frequency": 60}}}}'.format(mock_time.return_value))
        self._enrichment_kv.set('asn_api:namespace03',
                                '{{"data": {{"asn_data03": {{"data": 303}}}}, '
                                '"metadata": {{"_enrichment_group_creation_timestamp": {:.5f}, '
                                '"_enrichment_ttl": 300, "_execute_frequency": 60}}}}'.format(mock_time.return_value))
        self._enrichment_kv.set('enrichment:asn_api:namespace05', 'bad_data')

    def test_enrichment_cache(self):
        """Test enrichment resource attributes"""
        with self.assertRaises(AssertionError):
            PanoptesEnrichmentCache('non_panoptes_context', self._plugin_conf, self._panoptes_resource)

        with self.assertRaises(AssertionError):
            PanoptesEnrichmentCache(self._panoptes_context, 'non_plugin_conf', self._panoptes_resource)

        with self.assertRaises(AssertionError):
            PanoptesEnrichmentCache(self._panoptes_context, self._plugin_conf, 'non_panoptes_resource')

        #  Test with bad key-value store
        mock_panoptes_enrichment_cache_key_value_store = Mock(side_effect=Exception)
        with patch('yahoo_panoptes.framework.enrichment.PanoptesEnrichmentCacheKeyValueStore',
                   mock_panoptes_enrichment_cache_key_value_store):
            with self.assertRaises(Exception):
                PanoptesEnrichmentCache(self._panoptes_context, self._plugin_conf, self._panoptes_resource)

    def test_enrichment_cache_preload01(self):
        """Test enrichment resource with preload conf self:test_namespace"""

        result = '{{"test_resource_id": {{"test_namespace": {{"heartbeat": {{"timestamp": {}}}}}}}}}'.format(
            int(mock_time.return_value))

        enrichment_cache = PanoptesEnrichmentCache(self._panoptes_context,
                                                   self._plugin_conf,
                                                   self._panoptes_resource)

        self.assertEqual(ordered(enrichment_cache.__dict__['_enrichment_data']),
                         ordered(json.loads(result)))

    def test_enrichment_cache_preload02(self):
        """Test enrichment resource with preload conf"""
        plugin_conf = {'Core': {'name': 'Test Plugin 01'},
                       'enrichment': {'preload': 'self:test_namespace, '
                                                 'asn_api:*, '
                                                 'test_resource_id01:interface, '
                                                 'no_data_resource_id:test_namespace, '
                                                 'test_resource_id01:no_data_namespace'}
                       }

        results01 = """{{"asn_api": {{"namespace02": {{"asn_data02": {{"data": 202}}}}, "namespace03": {{"asn_data03":
        {{"data": 303}}}}, "namespace01": {{"asn_data01": {{"data": 101}}}}}}, "test_resource_id": {{"test_namespace":
        {{"heartbeat": {{"timestamp": {}}}}}}}, "test_resource_id01": {{"interface": {{"1226": {{"ifType": "l3ipvlan",
        "ifAlias": "<not set>", "ifPhysAddress": "00:1f:a0:13:8c:16", "ifDescr": "Virtual Ethernet 226", "ifName":
        "Virtual Ethernet 226", "ifMtu": "1500"}}, "1209": {{"ifType": "l3ipvlan", "ifAlias": "<not set>",
        "ifPhysAddress": "00:1f:a0:13:8c:15", "ifDescr": "Virtual Ethernet 209", "ifName": "Virtual Ethernet 209",
        "ifMtu": "1500"}}}}}}}}""".format(int(mock_time.return_value))

        enrichment_cache = PanoptesEnrichmentCache(self._panoptes_context,
                                                   plugin_conf,
                                                   self._panoptes_resource)

        self.assertEqual(ordered(enrichment_cache.__dict__['_enrichment_data']),
                         ordered(json.loads(results01)))

        #  Test an error while scanning kv store is handled correctly
        mock_find_keys = Mock(side_effect=Exception)
        mock_logger = MockLogger()
        with patch('yahoo_panoptes.framework.enrichment.PanoptesEnrichmentCacheKeyValueStore.find_keys',
                   mock_find_keys):
            with patch('yahoo_panoptes.framework.enrichment.PanoptesContext.logger', mock_logger):
                PanoptesEnrichmentCache(self._panoptes_context, plugin_conf, self._panoptes_resource)
                self.assertEqual(mock_logger.error.call_count, 3)

        mock_kv_store_get = Mock(side_effect=IOError)
        with patch('yahoo_panoptes.framework.enrichment.PanoptesEnrichmentCacheKeyValueStore.get',
                   mock_kv_store_get):
            with self.assertRaises(IOError):
                PanoptesEnrichmentCache(self._panoptes_context, plugin_conf, self._panoptes_resource)

    def test_enrichment_cache_parse_conf(self):
        """Test enrichment resource parse conf"""

        result01 = [('asn_api', '*'), ('self', 'test_namespace'), ('test_resource_id01', 'interface')]

        plugin_conf = {'Core': {'name': 'Test Plugin 01'},
                       'enrichment': {'preload': 'self:test_namespace, asn_api:*, test_resource_id01:interface'}
                       }

        enrichment_cache = PanoptesEnrichmentCache(self._panoptes_context,
                                                   plugin_conf,
                                                   self._panoptes_resource)
        self.assertEqual(sorted(result01), sorted(enrichment_cache.__dict__['_preload_conf']))

        plugin_conf_with_dup = {'Core': {'name': 'Test Plugin 01'},
                                'enrichment': {'preload': 'self:test_namespace, asn_api:*, '
                                                          'test_resource_id01:interface, asn_api:*, '
                                                          'test_resource_id01:interface'}}

        enrichment_cache = PanoptesEnrichmentCache(self._panoptes_context,
                                                   plugin_conf_with_dup,
                                                   self._panoptes_resource)
        self.assertEqual(sorted(result01), sorted(enrichment_cache.__dict__['_preload_conf']))

    def test_enrichment_cache_parse_conf_bad(self):
        """Test enrichment resource parse bad conf"""

        plugin_conf01 = {'Core': {'name': 'Test Plugin 01'},
                         'enrichment': {}
                         }
        with self.assertRaises(TypeError):
            PanoptesEnrichmentCache(self._panoptes_context, plugin_conf01, self._panoptes_resource)

        plugin_conf02 = {'Core': {'name': 'Test Plugin 01'},
                         'enrichment': {'preload': {}}
                         }
        with self.assertRaises(TypeError):
            PanoptesEnrichmentCache(self._panoptes_context, plugin_conf02, self._panoptes_resource)

        plugin_conf03 = {'Core': {'name': 'Test Plugin 01'},
                         'enrichment': {'preload': {'self:test_namespace, asn_api:'}}
                         }
        with self.assertRaises(TypeError):
            PanoptesEnrichmentCache(self._panoptes_context, plugin_conf03, self._panoptes_resource)

        plugin_conf04 = {'Core': {'name': 'Test Plugin 01'},
                         'enrichment': {'preload': {'self:test_namespace, :'}}
                         }
        with self.assertRaises(TypeError):
            PanoptesEnrichmentCache(self._panoptes_context, plugin_conf04, self._panoptes_resource)

        plugin_conf05 = {'Core': {'name': 'Test Plugin 01'},
                         'enrichment': {'preload': {'self:test_namespace, '}}
                         }
        with self.assertRaises(TypeError):
            PanoptesEnrichmentCache(self._panoptes_context, plugin_conf05, self._panoptes_resource)

        plugin_conf06 = {'Core': {'name': 'Test Plugin 01'},
                         'enrichment': {'preload': {'self:test_namespace, :namespace01'}}
                         }
        with self.assertRaises(TypeError):
            PanoptesEnrichmentCache(self._panoptes_context, plugin_conf06, self._panoptes_resource)

    def test_enrichment_cache_get_enrichment(self):
        """Test enrichment resource get"""
        plugin_conf = {'Core': {'name': 'Test Plugin 01'},
                       'enrichment': {'preload': 'self:test_namespace, '
                                                 'asn_api:*, '
                                                 'test_resource_id01:interface, '
                                                 'no_data_resource_id:test_namespace, '
                                                 'test_resource_id01:no_data_namespace'}
                       }

        enrichment_cache = PanoptesEnrichmentCache(self._panoptes_context,
                                                   plugin_conf,
                                                   self._panoptes_resource)

        result01 = {u'heartbeat': {u'timestamp': int(mock_time.return_value)}}
        enrichment_data = enrichment_cache.get_enrichment('test_resource_id', 'test_namespace')
        self.assertEqual(sorted(list(enrichment_data)), sorted(result01))

        enrichment_data = enrichment_cache.get_enrichment('self', 'test_namespace')
        self.assertEqual(sorted(list(enrichment_data)), sorted(result01))

        result02 = {u'1209': {u'ifType': u'l3ipvlan', u'ifAlias': u'<not set>', u'ifPhysAddress': u'00:1f:a0:13:8c:15',
                              u'ifDescr': u'Virtual Ethernet 209', u'ifName': u'Virtual Ethernet 209',
                              u'ifMtu': u'1500'},
                    u'1226': {u'ifType': u'l3ipvlan', u'ifAlias': u'<not set>', u'ifPhysAddress': u'00:1f:a0:13:8c:16',
                              u'ifDescr': u'Virtual Ethernet 226', u'ifName': u'Virtual Ethernet 226',
                              u'ifMtu': u'1500'}}

        enrichment_data = enrichment_cache.get_enrichment('test_resource_id01', 'interface')
        self.assertEqual(sorted(list(enrichment_data)), sorted(result02))

        with self.assertRaises(PanoptesEnrichmentCacheError):
            enrichment_cache.get_enrichment('no_data_resource_id', 'test_namespace')

        with self.assertRaises(PanoptesEnrichmentCacheError):
            enrichment_cache.get_enrichment('test_resource_id01', 'no_data_namespace')

        with self.assertRaises(AssertionError):
            enrichment_cache.get_enrichment('test_resource_id01', '')

        with self.assertRaises(AssertionError):
            enrichment_cache.get_enrichment('', 'test_namespace')

        with self.assertRaises(AssertionError):
            enrichment_cache.get_enrichment('', '*')

        # Test fallback preload
        result03 = {u'2209': {u'ifType': u'l3ipvlan', u'ifAlias': u'<not set>', u'ifPhysAddress': u'00:1f:a0:13:8c:15',
                              u'ifDescr': u'Virtual Ethernet 309', u'ifName': u'Virtual Ethernet 309',
                              u'ifMtu': u'1500'},
                    u'2226': {u'ifType': u'l3ipvlan', u'ifAlias': u'<not set>', u'ifPhysAddress': u'00:1f:a0:13:8c:16',
                              u'ifDescr': u'Virtual Ethernet 326', u'ifName': u'Virtual Ethernet 326',
                              u'ifMtu': u'1500'}}

        enrichment_data03 = enrichment_cache.get_enrichment('test_resource_id02', 'interface')
        self.assertEqual(sorted(list(enrichment_data03)), sorted(result03))

    def test_enrichment_cache_get_enrichment_value(self):
        """Test enrichment resource get_enrichment_value"""
        plugin_conf = {'Core': {'name': 'Test Plugin 01'},
                       'enrichment': {'preload': 'self:test_namespace, '
                                                 'asn_api:*, '
                                                 'test_resource_id01:interface, '
                                                 'no_data_resource_id:test_namespace, '
                                                 'test_resource_id01:no_data_namespace'}
                       }

        enrichment_cache = PanoptesEnrichmentCache(self._panoptes_context,
                                                   plugin_conf,
                                                   self._panoptes_resource)

        result01 = {u'timestamp': int(mock_time.return_value)}
        enrich_data = enrichment_cache.get_enrichment_value('self', 'test_namespace', 'heartbeat')
        self.assertEqual(sorted(result01), sorted(enrich_data))

        enrich_data = enrichment_cache.get_enrichment_value('test_resource_id', 'test_namespace', 'heartbeat')
        self.assertEqual(sorted(result01), sorted(enrich_data))

        result02 = {u'data': 101}
        enrich_data = enrichment_cache.get_enrichment_value('asn_api', 'namespace01', 'asn_data01')
        self.assertEqual(sorted(result02), sorted(enrich_data))

        result03 = {u'ifType': u'l3ipvlan', u'ifAlias': u'<not set>', u'ifPhysAddress': u'00:1f:a0:13:8c:15',
                    u'ifDescr': u'Virtual Ethernet 209', u'ifName': u'Virtual Ethernet 209', u'ifMtu': u'1500'}
        enrich_data = enrichment_cache.get_enrichment_value('test_resource_id01', 'interface', '1209')
        self.assertEqual(sorted(result03), sorted(enrich_data))

        with self.assertRaises(PanoptesEnrichmentCacheError):
            enrichment_cache.get_enrichment_value('no_data_resource_id', 'test_namespace', 'no_data')

        with self.assertRaises(AssertionError):
            enrichment_cache.get_enrichment_value('no_data_resource_id', 'test_namespace', '')

        with self.assertRaises(AssertionError):
            enrichment_cache.get_enrichment_value('no_data_resource_id', '', 'no_data')

        with self.assertRaises(AssertionError):
            enrichment_cache.get_enrichment_value('', 'test_namespace', 'no_data')

        # Test fallback preload
        result04 = {u'ifType': u'l3ipvlan', u'ifAlias': u'<not set>', u'ifPhysAddress': u'00:1f:a0:13:8c:15',
                    u'ifDescr': u'Virtual Ethernet 309', u'ifName': u'Virtual Ethernet 309', u'ifMtu': u'1500'}

        enrich_data = enrichment_cache.get_enrichment_value('test_resource_id02', 'interface', '2209')
        self.assertEqual(sorted(result04), sorted(enrich_data))

    def test_enrichment_cache_get_enrichment_keys(self):
        """Test enrichment resource get_enrichment_keys"""
        plugin_conf = {'Core': {'name': 'Test Plugin 01'},
                       'enrichment': {'preload': 'self:test_namespace, '
                                                 'asn_api:*, '
                                                 'test_resource_id01:interface, '
                                                 'no_data_resource_id:test_namespace, '
                                                 'test_resource_id01:no_data_namespace'}
                       }

        enrichment_cache = PanoptesEnrichmentCache(self._panoptes_context,
                                                   plugin_conf,
                                                   self._panoptes_resource)

        result01 = [u'1226', u'1209']
        enrichment_data = enrichment_cache.get_enrichment_keys('test_resource_id01', 'interface')
        self.assertListEqual(enrichment_data, result01)

        with self.assertRaises(PanoptesEnrichmentCacheError):
            enrichment_cache.get_enrichment_keys('test_resource_id01', 'no_data_namespace')

        result02 = [u'heartbeat']
        enrichment_data = enrichment_cache.get_enrichment_keys('test_resource_id', 'test_namespace')
        self.assertListEqual(enrichment_data, result02)

        enrichment_data = enrichment_cache.get_enrichment_keys('self', 'test_namespace')
        self.assertListEqual(enrichment_data, result02)

        # Test fallback preload
        result03 = [u'2226', u'2209']
        enrichment_data03 = enrichment_cache.get_enrichment_keys('test_resource_id02', 'interface')
        self.assertListEqual(enrichment_data03, result03)



"""
Copyright 2018, Oath Inc.
Licensed under the terms of the Apache 2.0 license. See LICENSE file in project root for terms.

This module implements a 'runner' that cna take a given plugin name and type and execute it, validate and return the
results to a callback function

It also updates metadata like the plugin's last execution time and last results time
"""
import gc
import time
import weakref

from ..validators import PanoptesValidators
from ..context import PanoptesContext
from ..enrichment import PanoptesEnrichmentCacheError, PanoptesEnrichmentCache
from ..plugins.context import PanoptesPluginContext, PanoptesPluginWithEnrichmentContext
from ..plugins.manager import PanoptesPluginManager
from ..plugins.panoptes_base_plugin import PanoptesBasePluginValidators, PanoptesPluginInfoValidators
from ..utilities.key_value_store import PanoptesKeyValueStoreValidators


class PanoptesPluginRunner(object):
    """
    This class implements a 'runner' for a given plugin

    Args:
        plugin_name (str): The name of the plugin to run
        plugin_name (str): The type of the plugin
        plugin_name (PanoptesBasePlugin): The plugin class
        plugin_info_class (PanoptesPluginInfo): The plugin info class
        plugin_data (object): An optional arbitrary data object to be passed to the plugin
        panoptes_context (PanoptesContext): The Panoptes Context associated with the Plugin Agent executing the plugin
        plugin_agent_kv_store_class (class): The class of the Plugin Agent's KV store
        plugin_kv_store_class (class): The class of the Plugin's KV store
        plugin_kv_store_class (class): The class of the Plugin's secrets store
        plugin_logger_name (str): The name of the logger that the plugin should use
        plugin_result_class (class): The class of the results expected from the plugin
        results_callback (callable): The function that should called with the results generated from the plugin
    """

    def __init__(self, plugin_name, plugin_type, plugin_class, plugin_info_class, plugin_data, panoptes_context,
                 plugin_agent_kv_store_class, plugin_kv_store_class, plugin_secrets_store_class,
                 plugin_logger_name, plugin_result_class, results_callback):
        assert PanoptesValidators.valid_nonempty_string(plugin_name), 'plugin_name must be a non-empty string'
        assert PanoptesValidators.valid_nonempty_string(plugin_type), 'plugin_type must be a non-empty string'
        assert PanoptesBasePluginValidators.valid_plugin_class(
                plugin_class), 'plugin_class must be instance of PanoptesBasePlugin'
        assert PanoptesPluginInfoValidators.valid_plugin_info_class(
                plugin_info_class), 'plugin_info_class must be instance of PanoptesPluginInfo'
        assert PanoptesValidators.valid_hashable_object(plugin_data), 'plugin_data must be a valid hashable object'
        assert PanoptesKeyValueStoreValidators.valid_kv_store_class(
                plugin_kv_store_class), 'plugin_kv_store_class must be a subclass of PanoptesKeyValueStore'
        assert PanoptesKeyValueStoreValidators.valid_kv_store_class(
                plugin_secrets_store_class), 'plugin_secrets_store_class must be a subclass of PanoptesKeyValueStore'
        assert PanoptesValidators.valid_nonempty_string(plugin_logger_name), 'plugin_logger_name must be a non-empty '\
                                                                             'string'
        assert PanoptesValidators.valid_callback(results_callback), 'plugin_callback must be a callable'
        self._plugin_name = plugin_name
        self._plugin_type = plugin_type
        self._plugin_class = plugin_class
        self._plugin_info_class = plugin_info_class
        self._plugin_data = plugin_data
        self._panoptes_context = panoptes_context
        self._plugin_agent_kv_store_class = plugin_agent_kv_store_class
        self._plugin_secrets_store_class = plugin_secrets_store_class
        self._plugin_kv_store_class = plugin_kv_store_class
        self._plugin_logger_name = plugin_logger_name
        self._logger = self._panoptes_context.logger
        self._plugin_result_class = plugin_result_class
        self._results_callback = weakref.proxy(results_callback)

    def info(self, plugin, message):
        self._logger.info('[{}:{}] [{}] {}'.format(plugin.name, plugin.signature, str(plugin.data), message))

    def warn(self, plugin, message):
        self._logger.warn('[{}:{}] [{}] {}'.format(plugin.name, plugin.signature, str(plugin.data), message))

    def error(self, plugin, message, exception):
        self._logger.error(
                '[{}:{}] [{}] {}: {}'.format(plugin.name, plugin.signature, str(plugin.data), message, repr(exception)))

    def exception(self, plugin, message):
        self._logger.exception('[{}:{}] [{}] {}:'.format(plugin.name, plugin.signature, str(plugin.data), message))

    def _get_context(self, plugin):
        return PanoptesPluginContext(panoptes_context=self._panoptes_context,
                                     logger_name=self._plugin_logger_name,
                                     config=plugin.config,
                                     key_value_store=self._panoptes_context.get_kv_store(
                                         self._plugin_kv_store_class),
                                     secrets_store=self._panoptes_context.get_kv_store(
                                         self._plugin_secrets_store_class),
                                     data=self._plugin_data)

    def execute_plugin(self):
        """
        This method loads, executes and returns the results of the plugin

        The steps involved are:
          * Create a PanoptesPluginManager for the type of plugins we are interested in and then try and load the \
        named plugin
          * Check if the plugin should be executed right now based on the last execution and results time and the \
          configured execution and results caching time
          * setup a PluginContext and attempt to take a lock on the plugin name and it's signature \
        (which is hash of the plugin's config and data)
          * Try and run the plugin
          * If the plugin execution returns without errors, then update the last execution time
          * Pass the results obtained to the callback function
          * If the callback function succeeds, then update the last successful run time

        Returns:
            None
        """
        logger = self._logger
        config = self._panoptes_context.config_dict
        utc_epoch = int(time.time())

        logger.info('Attempting to execute plugin "%s"' % self._plugin_name)

        try:
            plugin_manager = PanoptesPluginManager(plugin_type=self._plugin_type,
                                                   plugin_class=self._plugin_class,
                                                   plugin_info_class=self._plugin_info_class,
                                                   plugin_data=self._plugin_data,
                                                   panoptes_context=self._panoptes_context,
                                                   kv_store_class=self._plugin_agent_kv_store_class)

            plugin = plugin_manager.getPluginByName(name=self._plugin_name, category=self._plugin_type)
        except Exception as e:
            logger.error('Error trying to load plugin "%s": %s' % (self._plugin_name, repr(e)))
            return

        if not plugin:
            logger.warn('No plugin named "%s" found in "%s"' % (
                self._plugin_name, str(config[self._plugin_type]['plugins_paths'])))
            return

        if not plugin.execute_now:
            return

        try:
            plugin_context = self._get_context(plugin)
        except:
            self.exception(plugin, 'Could not setup context for plugin')
            return

        self.info(plugin, 'Attempting to get lock for plugin "%s"' % self._plugin_name)

        try:
            lock = plugin.lock
        except:
            self.exception(plugin, 'Error in acquiring lock')
            return

        if lock is None:
            return

        self.info(plugin, 'Acquired lock')

        self.info(plugin,
                  'Going to run plugin "{}", version "{}", which last executed at {} (UTC) ({} seconds ago) and '
                  'last produced results at {} (UTC) ({} seconds ago), module mtime {} (UTC), config mtime {} ('
                  'UTC)'.format(
                          plugin.name, plugin.version, plugin.last_executed,
                          plugin.last_executed_age,
                          plugin.last_results,
                          plugin.last_results_age, plugin.moduleMtime, plugin.configMtime))

        results = None

        plugin_start_time = time.time()
        try:
            results = plugin.plugin_object.run(plugin_context)
        except:
            self.exception(plugin, 'Failed to execute plugin')
        finally:
            plugin_end_time = time.time()
            self.info(plugin, 'Ran in {:0.2f} seconds'.format(plugin_end_time - plugin_start_time))
            try:
                lock.release()
            except:
                self.exception(plugin, 'Failed to release lock for plugin')
            else:
                self.info(plugin, 'Released lock')

        if results is None:
            self.warn(plugin, 'Plugin did not return any results')
            return
        elif not isinstance(results, self._plugin_result_class):
            logger.warn(plugin, 'Plugin returned an unexpected result type: "{}"'.format(type(results).__name__))
            return

        plugin.last_executed = utc_epoch

        self.info(plugin, 'Plugin returned a result set with {} members'.format(len(results)))

        if len(results) > 0:
            # Non-empty result set - send the results to the callback function
            callback_start_time = time.time()
            try:
                self._results_callback(self._panoptes_context, results, plugin)
            except:
                self.exception(plugin, 'Results callback function failed')
                return
            finally:
                callback_end_time = time.time()
                self.info(plugin,
                          'Callback function ran in {:0.2f} seconds'.format(callback_end_time - callback_start_time))

            # If the callback was successful, then set the last results time
            # The logic behind this is: in case the callback fails, then the plugin should be re-executed again after
            # the plugin execute_frequency seconds - the execution should not be preempted by the results caching logic,
            # which depends on the last results time in the KV store

            plugin.last_results = utc_epoch

        gc_start_time = time.time()
        gc.collect()
        gc_end_time = time.time()

        logger.info(
            'GC took %.2f seconds. There are %d garbage objects.' % (gc_end_time - gc_start_time, len(gc.garbage)))


class PanoptesPluginWithEnrichmentRunner(PanoptesPluginRunner):
    def _get_context(self, plugin):

        self._enrichment = None

        if plugin.config.get('enrichment'):
            try:
                self._enrichment = PanoptesEnrichmentCache(self._panoptes_context, plugin.config, self._plugin_data)
            except Exception as e:
                raise PanoptesEnrichmentCacheError('Error while creating PanoptesEnrichmentResource object for plugin '
                                                   '{}: {}, skipping run'.format(plugin.name, repr(e)))

            if self._enrichment is None:
                raise PanoptesEnrichmentCacheError('No enrichments found for plugin {} (configured {}), '
                                                   'skipping run'.format(plugin.name, plugin.config.get('enrichment')))

        return PanoptesPluginWithEnrichmentContext(panoptes_context=self._panoptes_context,
                                                   logger_name=self._plugin_logger_name,
                                                   config=plugin.config,
                                                   key_value_store=self._panoptes_context.get_kv_store(
                                                       self._plugin_kv_store_class),
                                                   secrets_store=self._panoptes_context.get_kv_store(
                                                       self._plugin_secrets_store_class),
                                                   data=self._plugin_data,
                                                   enrichment=self._enrichment)



"""
Copyright 2018, Oath Inc.
Licensed under the terms of the Apache 2.0 license. See LICENSE file in project root for terms.

This module implements Heartbeat EnrichmentGroup defined with schema validator
"""

from ...framework.enrichment import PanoptesEnrichmentSchemaValidator, PanoptesEnrichmentGroup


HEARTBEAT_SCHEMA_NAMESPACE = 'heartbeat_ns'


class PanoptesHeartbeatEnrichmentSchemaValidator(PanoptesEnrichmentSchemaValidator):
    schema = {
        'enrichment_label': {
            'type': 'dict',
            'schema': {
                'timestamp': {'type': 'float'},
            }
        }
    }


class PanoptesHeartbeatEnrichmentGroup(PanoptesEnrichmentGroup):
    def __init__(self, enrichment_ttl, execute_frequency):
        super(PanoptesHeartbeatEnrichmentGroup, self).__init__(
            namespace=HEARTBEAT_SCHEMA_NAMESPACE,
            schema_validator=PanoptesHeartbeatEnrichmentSchemaValidator(),
            enrichment_ttl=enrichment_ttl,
            execute_frequency=execute_frequency)



"""
Copyright 2018, Oath Inc.
Licensed under the terms of the Apache 2.0 license. See LICENSE file in project root for terms.


This module implements Metrics EnrichmentGroup defined with schema validator
"""

from ....framework.enrichment import PanoptesEnrichmentSchemaValidator, PanoptesEnrichmentGroup


class PanoptesGenericSNMPMetricsEnrichmentSchemaValidator(PanoptesEnrichmentSchemaValidator):
    schema = {
        'enrichment_label': {
            'type': 'dict',
            'schema': {
                'oids': {
                    'type': 'dict', 'required': True
                },
                'metrics_groups': {
                    'type': 'list',
                    'required': True,
                    'schema': {
                        'type': 'dict',
                        'schema': {
                            'group_name': {
                                'type': 'string',
                                'required': True
                            },
                            'dimensions': {
                                'type': 'dict', 'required': False
                            },
                            'metrics': {
                                'type': 'dict', 'required': True
                            }
                        }
                    }
                }
            }
        }
    }


class PanoptesGenericSNMPMetricsEnrichmentGroup(PanoptesEnrichmentGroup):
    METRICS_SCHEMA_NAMESPACE = 'metrics'

    def __init__(self, enrichment_ttl, execute_frequency):
        super(PanoptesGenericSNMPMetricsEnrichmentGroup, self).__init__(
            namespace=self.METRICS_SCHEMA_NAMESPACE,
            schema_validator=PanoptesGenericSNMPMetricsEnrichmentSchemaValidator(),
            enrichment_ttl=enrichment_ttl,
            execute_frequency=execute_frequency)



"""
Copyright 2018, Oath Inc.
Licensed under the terms of the Apache 2.0 license. See LICENSE file in project root for terms.

This module implements a context object that is passed to each plugin during execution

The context of plugin contains it's configuration, the logger it should use, the key/value store it can use and an
optional, arbitrary data object to be passed to the plugin
"""
from ..context import PanoptesContext, PanoptesContextValidators
from ..utilities.key_value_store import PanoptesKeyValueStore


class PanoptesPluginContext(object):
    """
    This class defines the context which is passed to each plugin instance when it is executed

    Args:
        panoptes_context(PanoptesContext): The Panoptes Context associated with the Plugin Agent that
        executed the plugin
        logger_name(str): The name of the logger to use
        config(dict): The plugin configuration parsed from the .panoptes-plugin info file associated with the plugin
        key_value_store(PanoptesKeyValueStore): The Key/Value store to use
        secrets_store(PanoptesKeyValueStore): The secrets key/value store to use
        data(object): An optional data object which would be passed through to the plugin being executed
    """

    def __init__(self, panoptes_context, logger_name, config, key_value_store, secrets_store, data=None):
        assert PanoptesContextValidators.valid_panoptes_context(
            panoptes_context), 'panoptes_context must be an instance of PanoptesContext'
        assert logger_name and isinstance(logger_name, str), 'logger_name must be a non-empty str'
        assert config is None or isinstance(config, dict), 'config must be a dict'
        assert isinstance(secrets_store,
                          PanoptesKeyValueStore), 'secrets_store must be an instance of PanoptesKeyValueStore'
        assert isinstance(key_value_store,
                          PanoptesKeyValueStore), 'key_value_store must be an instance of PanoptesKeyValueStore'
        self._panoptes_context = panoptes_context
        self._logger = panoptes_context.logger.getChild(logger_name)
        self._config = config
        self._kv_store = key_value_store
        self._secrets_store = secrets_store
        self._data = data
        self._sites = panoptes_context.config_object.sites
        self._snmp_defaults = panoptes_context.config_object.snmp_defaults

    @property
    def logger(self):
        """
        The logger being used by the plugin

        Returns:
            logging.logger: The logger being used by the plugin

        """
        return self._logger

    @property
    def config(self):
        """
        The config being used by the plugin

        Returns:
            dict: The config being used by the plugin
        """
        return self._config

    @property
    def kv(self):
        """
        The Key/Value store being used by the plugin

        Returns:
            PanoptesKeyValueStore: The Key/Value store being used by the plugin
        """
        return self._kv_store

    @property
    def secrets(self):
        """
        The secrets store being used by the plugin

        Returns:
            PanoptesKeyValueStore: The secrets store being used by the plugin
        """
        return self._secrets_store

    @property
    def data(self):
        """
        The data object passed to the plugin

        Returns:
            object:  The data object passed to the plugin
        """
        return self._data

    @property
    def sites(self):
        """
        The set of sites used by the plugin

        Returns:
            set(str):  Set of sites used by the plugin
        """
        return self._sites

    @property
    def snmp(self):
        """
        The SNMP configuration from the site configuration file

        Returns:
            dict: A dictionary containing SNMP configuration
        """
        return self._snmp_defaults


class PanoptesPluginWithEnrichmentContext(PanoptesPluginContext):
    def __init__(self, panoptes_context, logger_name, config, key_value_store,
                 secrets_store, data=None, enrichment=None):
        super(PanoptesPluginWithEnrichmentContext, self).__init__(panoptes_context,
                                                                  logger_name,
                                                                  config,
                                                                  key_value_store,
                                                                  secrets_store,
                                                                  data=data)
        self._enrichment = enrichment

    @property
    def enrichment(self):
        """
        The enrichment object passed to the plugin

        Returns:
            object:  The enrichment object passed to the plugin
        """
        return self._enrichment



import re

from kazoo.client import KazooState
from kazoo.exceptions import LockTimeout

from ..validators import PanoptesValidators
from ..context import PanoptesContextValidators
from .helpers import get_calling_module_name


class PanoptesLock(object):
    def __init__(self, context, path, timeout, retries=1, identifier=None):
        """
        Creates and maintains state for a lock

        Args:
            path (str): A '/' separated path for the lock
            timeout (int): in seconds. Must be a positive integer
            retries (int): how many times to try before giving up. Zero implies try forever
            identifier (str): Name to use for this lock contender. This can be useful for querying \
            to see who the current lock contenders are

        Returns:
            PanoptesLock: lock
        """
        assert PanoptesContextValidators.valid_panoptes_context(context), 'context must be a valid PanoptesContext'
        assert isinstance(path, str) and re.search("^/\S+", path), 'path must be a non-empty string that begins with /'
        assert PanoptesValidators.valid_nonzero_integer(timeout), 'timeout must be a positive integer'
        assert PanoptesValidators.valid_positive_integer(retries), 'retries must be a non-negative integer'
        assert PanoptesValidators.valid_nonempty_string(identifier), 'identifier must be a non-empty string'

        self._context = context
        self._logger = self._context.logger
        self._path = path
        self._timeout = timeout
        self._retries = retries
        self._identifier = identifier
        self._lock = None
        self._locked = False
        self._calling_module = get_calling_module_name(3)

        self._get_lock()

    def __str__(self):
        return 'calling module: {}, path={}, timeout={}, retries={}, identifier={}'.format(
                self._calling_module, self._path, self._timeout, self._retries, self._identifier)

    @property
    def locked(self):
        return self._locked

    def release(self):
        if self._lock:
            self._logger.info('Releasing lock for {}'.format(str(self)))
            self._lock.release()

        self._locked = False
        self._lock = None

    def _get_lock(self):
        """
        A wrapper around the Kazoo library's lock. On successful acquisition of the lock, sets self._lock and
        self._locked

        Returns:
            None
        """
        logger = self._logger

        logger.info('Creating lock for {}'.format(str(self)))

        try:
            lock = self._context.zookeeper_client.Lock(self._path, self._identifier)
        except:
            logger.exception('Failed to create lock object')
            return

        tries = 0
        while (self._retries == 0) or (tries < self._retries):
            tries += 1
            logger.info('Trying to acquire lock for {}. Other contenders: {}'.format(str(self), lock.contenders()))
            try:
                lock.acquire(timeout=self._timeout)
            except LockTimeout:
                logger.info('Timed out after {} seconds trying to acquire lock for {}'.format(self._timeout,
                                                                                              str(self)))
            except Exception as e:
                logger.info('Error in acquiring lock for {}: {}'.format(str(self), repr(e)))

            if lock.is_acquired:
                break

        if not lock.is_acquired:
            logger.warn('Failed to acquire lock for {} after {} tries'.format(str(self), tries))
        else:
            logger.info('Lock acquired for {}. Other contenders: {}'.format(str(self), lock.contenders()))
            self._locked = True
            self._context.zookeeper_client.add_listener(self._lock_listener)
            self._lock = lock

    def _release_and_reacquire(self):
        if self._lock:
            self.release()
            self._get_lock()

    def _lock_listener(self, state):
        """
        Listener to handle ZK disconnection/reconnection. Since I don't know of safe way to check if a lock is still in
        ZK after a reconnect, we simply release the lock and try and re-acquire it.

        Args:
            state (kazoo.client.KazooState): The state of the ZK connection

        Returns:
            None
        """

        if state in [KazooState.LOST, KazooState.SUSPENDED]:
            self._logger.warn('Disconnected from Zookeeper, waiting to reconnect lock for {}'.format(str(self)))
            self._locked = False
        elif state == KazooState.CONNECTED:
            self._logger.warn(
                    'Reconnected to Zookeeper, trying to release and re-acquire lock for {}'.format(str(self)))
            self._context.zookeeper_client.handler.spawn(self._release_and_reacquire)
        else:
            self._logger.warn('Got unknown state "{}" from Zookeeper'.format(state))



"""
Copyright 2018, Oath Inc.
Licensed under the terms of the Apache 2.0 license. See LICENSE file in project root for terms.
"""
import os
import time
import unittest

from mock import patch, MagicMock
from yapsy.PluginInfo import PluginInfo

from yahoo_panoptes.framework.plugins.panoptes_base_plugin import PanoptesPluginInfo, PanoptesPluginInfoValidators, \
    PanoptesPluginConfigurationError, PanoptesBasePluginValidators, PanoptesBasePlugin
from yahoo_panoptes.polling.polling_plugin import PanoptesPollingPluginInfo
from yahoo_panoptes.framework.resources import PanoptesResource, PanoptesContext
from yahoo_panoptes.framework.utilities.helpers import get_module_mtime

from .test_framework import PanoptesTestKeyValueStore, panoptes_mock_kazoo_client, panoptes_mock_redis_strict_client
from .helpers import get_test_conf_file

_TIMESTAMP = round(time.time(), 5)
_LAST_EXECUTED_TEST_VALUE = 1458947997
_LAST_RESULTS_TEST_VALUE = 1458948005

mock_time = MagicMock(return_value=_TIMESTAMP)


class TestPanoptesBasePluginSubclass(PanoptesBasePlugin):
    def run(self, context):
        pass


class TestPanoptesBasePluginValidators(unittest.TestCase):
    def test_valid_plugin_class(self):
        self.assertFalse(PanoptesBasePluginValidators.valid_plugin_class(None))
        self.assertFalse(PanoptesBasePluginValidators.valid_plugin_class(PanoptesContext))
        self.assertTrue(PanoptesBasePluginValidators.valid_plugin_class(PanoptesBasePlugin))
        self.assertTrue(PanoptesBasePluginValidators.valid_plugin_class(TestPanoptesBasePluginSubclass))


class TestPanoptesPluginInfoValidators(unittest.TestCase):
    def test_valid_plugin_info_class(self):
        self.assertFalse(PanoptesPluginInfoValidators.valid_plugin_info_class(None))
        self.assertTrue(PanoptesPluginInfoValidators.valid_plugin_info_class(PanoptesPollingPluginInfo))
        self.assertTrue(PanoptesPluginInfoValidators.valid_plugin_info_class(PanoptesPluginInfo))
        self.assertFalse(PanoptesPluginInfoValidators.valid_plugin_info_class(PluginInfo))


def mock_metadata_kv_store():
    return MagicMock(side_effect=Exception)


class TestPanoptesPluginInfo(unittest.TestCase):
    @patch('redis.StrictRedis', panoptes_mock_redis_strict_client)
    @patch('kazoo.client.KazooClient', panoptes_mock_kazoo_client)
    def setUp(self):
        self.my_dir, self.panoptes_test_conf_file = get_test_conf_file()
        self._panoptes_context = PanoptesContext(self.panoptes_test_conf_file,
                                                 key_value_store_class_list=[PanoptesTestKeyValueStore],
                                                 create_message_producer=False, async_message_producer=False,
                                                 create_zookeeper_client=True)
        self._panoptes_resource = PanoptesResource(resource_site='test', resource_class='test',
                                                   resource_subclass='test',
                                                   resource_type='test', resource_id='test', resource_endpoint='test',
                                                   resource_plugin='test')

    def test_plugin_info_repr(self):
        panoptes_plugin_info = PanoptesPluginInfo("plugin_name", "path/to/plugin")
        panoptes_plugin_info.panoptes_context = self._panoptes_context
        panoptes_plugin_info.data = self._panoptes_resource
        panoptes_plugin_info.kv_store_class = PanoptesTestKeyValueStore
        panoptes_plugin_info.last_executed = _LAST_EXECUTED_TEST_VALUE
        panoptes_plugin_info.last_results = _LAST_RESULTS_TEST_VALUE

        repr_string = "PanoptesPluginInfo: Normalized name: plugin__name, Config file: None, " \
                      "Panoptes context: [PanoptesContext: KV Stores: [PanoptesTestKeyValueStore], " \
                      "Config: ConfigObj({'main': {'sites': ['local'], " \
                      "'plugins_extension': 'panoptes-plugin', 'plugins_skew': 1}, " \
                      "'log': {'config_file': 'tests/config_files/test_panoptes_logging.ini', 'rate': 1000, " \
                      "'per': 1, 'burst': 10000, " \
                      "'formatters': {'keys': ['root_log_format', 'log_file_format', 'discovery_plugins_format']}}, " \
                      "'redis': {'default': {'namespace': 'panoptes', 'shards': {'shard1': " \
                      "{'host': 'localhost', 'port': 6379, 'db': 0, 'password': '**'}}}}, " \
                      "'kafka': {'topic_key_delimiter': ':', 'topic_name_delimiter': '-', " \
                      "'brokers': {'broker1': {'host': 'localhost', 'port': 9092}}, " \
                      "'topics': {'metrics': {'raw_topic_name_suffix': 'metrics', " \
                      "'transformed_topic_name_suffix': 'processed'}}}, " \
                      "'zookeeper': {'connection_timeout': 30, 'servers': {'server1': " \
                      "{'host': 'localhost', 'port': 2181}}}, 'discovery': {" \
                      "'plugins_paths': ['tests/plugins/discovery'], 'plugin_scan_interval': 60, " \
                      "'celerybeat_max_loop_interval': 5}, " \
                      "'polling': {'plugins_paths': ['tests/plugins/polling'], 'plugin_scan_interval': 60, " \
                      "'celerybeat_max_loop_interval': 5}, " \
                      "'enrichment': {'plugins_paths': ['tests/plugins/enrichment'], " \
                      "'plugin_scan_interval': 60, 'celerybeat_max_loop_interval': 5}, " \
                      "'snmp': {'port': 10161, 'connection_factory_module': " \
                      "'yahoo_panoptes.framework.utilities.snmp.connection', 'connection_factory_class': " \
                      "'PanoptesSNMPConnectionFactory', 'community': '**', 'timeout': 5, 'retries': 1, " \
                      "'non_repeaters': 0, 'max_repetitions': 25, 'proxy_port': 10161, " \
                      "'community_string_key': 'snmp_community_string'}}), Redis pool set: False, " \
                      "Message producer set: False, Kafka client set: False, Zookeeper client set: False], " \
                      "KV store class: PanoptesTestKeyValueStore, Last executed timestamp: 1458947997, " \
                      "Last executed key: plugin_metadata:plugin__name:" \
                      "61547fbb304169f2a076016678bc9cca:last_executed, " \
                      "Last results timestamp: 1458948005, " \
                      "Last results key: plugin_metadata:plugin__name:" \
                      "61547fbb304169f2a076016678bc9cca:last_results, " \
                      "Data: Data object passed, " \
                      "Lock: Lock is set"
        self.assertEqual(repr(panoptes_plugin_info), repr_string)

    def test_plugin_info_moduleMtime(self):
        panoptes_plugin_info = PanoptesPluginInfo("plugin_name", "path/to/plugin")
        panoptes_plugin_info.path = self.my_dir

        self.assertEqual(panoptes_plugin_info.moduleMtime, get_module_mtime(self.my_dir))

        panoptes_plugin_info.path = 0
        with self.assertRaises(PanoptesPluginConfigurationError):
            panoptes_plugin_info.moduleMtime

    def test_plugin_info_configMtime(self):
        panoptes_plugin_info = PanoptesPluginInfo("plugin_name", "path/to/plugin")
        panoptes_plugin_info.config_filename = self.my_dir
        self.assertEqual(panoptes_plugin_info.configMtime, int(os.path.getmtime(self.my_dir)))

        panoptes_plugin_info.config_filename = "/non/existent/file"
        with self.assertRaises(PanoptesPluginConfigurationError):
            panoptes_plugin_info.configMtime

    def test_plugin_info_last_properties(self):
        panoptes_plugin_info = PanoptesPluginInfo("Test Polling Plugin",
                                                  "tests/plugins/polling/test")

        #  Test last_results and last_executed return 0 on exception.
        with self.assertRaises(AssertionError):
            panoptes_plugin_info.last_executed = "test"
        self.assertEqual(panoptes_plugin_info.last_executed, 0)

        with self.assertRaises(AssertionError):
            panoptes_plugin_info.last_results = "test"
        self.assertEqual(panoptes_plugin_info.last_results, 0)

    def test_plugin_info_properties(self):
        panoptes_plugin_info = PanoptesPluginInfo("Test Polling Plugin",
                                                  "tests/plugins/polling/test")

        with self.assertRaises(PanoptesPluginConfigurationError):
            panoptes_plugin_info.panoptes_context

        panoptes_plugin_info.panoptes_context = self._panoptes_context
        panoptes_plugin_info.kv_store_class = PanoptesTestKeyValueStore
        panoptes_plugin_info.last_executed = _LAST_EXECUTED_TEST_VALUE
        panoptes_plugin_info.config_filename = "tests/plugins/polling/test/plugin_polling_test.panoptes-plugin"

        #  Test results_cache_age, execute_frequency, last_executed return 0 on exception.
        self.assertEqual(panoptes_plugin_info.results_cache_age, 0)
        self.assertEqual(panoptes_plugin_info.execute_frequency, 0)

        panoptes_plugin_info.details.read(panoptes_plugin_info.config_filename)

        self.assertEqual(panoptes_plugin_info.results_cache_age, 100)
        self.assertEqual(panoptes_plugin_info.execute_frequency, 60)

        #  Test last_executed setter cannot be reset after being read
        self.assertEqual(panoptes_plugin_info.last_executed, _LAST_EXECUTED_TEST_VALUE)
        panoptes_plugin_info.last_executed = 1
        self.assertNotEqual(panoptes_plugin_info.last_executed, 1)
        self.assertEqual(panoptes_plugin_info.last_executed, _LAST_EXECUTED_TEST_VALUE)

        with patch("yahoo_panoptes.framework.plugins.panoptes_base_plugin.time.time", mock_time):
            self.assertEqual(panoptes_plugin_info.last_executed_age, int(_TIMESTAMP - _LAST_EXECUTED_TEST_VALUE))

    def test_plugin_info_last_results(self):
        panoptes_plugin_info = PanoptesPluginInfo("plugin_name", "path/to/plugin")
        panoptes_plugin_info.panoptes_context = self._panoptes_context
        panoptes_plugin_info.kv_store_class = PanoptesTestKeyValueStore

        panoptes_plugin_info.last_results = _LAST_RESULTS_TEST_VALUE
        self.assertEqual(panoptes_plugin_info.last_results, _LAST_RESULTS_TEST_VALUE)

        # Test value is cached.
        self.assertEqual(panoptes_plugin_info.last_results, _LAST_RESULTS_TEST_VALUE)

        #  Test last_results setter handles exception.
        with patch('yahoo_panoptes.framework.plugins.panoptes_base_plugin.PanoptesPluginInfo.metadata_kv_store',
                   mock_metadata_kv_store):
            panoptes_plugin_info.last_results = 1
            self.assertNotEqual(panoptes_plugin_info.last_results, 1)
            self.assertEqual(panoptes_plugin_info.last_results, _LAST_RESULTS_TEST_VALUE)

        #  Test last results age.
        with patch("yahoo_panoptes.framework.plugins.panoptes_base_plugin.time.time", mock_time):
            self.assertEqual(panoptes_plugin_info.last_results_age, int(_TIMESTAMP - _LAST_RESULTS_TEST_VALUE))

    def test_plugin_info_last_executed(self):
        panoptes_plugin_info = PanoptesPluginInfo("plugin_name", "path/to/plugin")
        panoptes_plugin_info.panoptes_context = self._panoptes_context
        panoptes_plugin_info.kv_store_class = PanoptesTestKeyValueStore

        #  Test last_executed setter handles exception.
        with patch('yahoo_panoptes.framework.plugins.panoptes_base_plugin.PanoptesPluginInfo.metadata_kv_store',
                   mock_metadata_kv_store):
            panoptes_plugin_info.last_executed = 1
            self.assertNotEqual(panoptes_plugin_info.last_executed, 1)
            self.assertEqual(panoptes_plugin_info.last_executed, 0)

    def test_plugin_info_execute_now(self):
        panoptes_plugin_info = PanoptesPluginInfo("plugin_name", "path/to/plugin")
        panoptes_plugin_info.panoptes_context = self._panoptes_context
        panoptes_plugin_info.kv_store_class = PanoptesTestKeyValueStore
        panoptes_plugin_info.config_filename = "tests/plugins/polling/test/plugin_polling_test.panoptes-plugin"
        panoptes_plugin_info.details.read(panoptes_plugin_info.config_filename)

        mock_moduleMtime = _TIMESTAMP - 1
        mock_configMtime = _TIMESTAMP - 2
        with patch('yahoo_panoptes.framework.plugins.panoptes_base_plugin.PanoptesPluginInfo.configMtime',
                   mock_configMtime):
            with patch('yahoo_panoptes.framework.plugins.panoptes_base_plugin.PanoptesPluginInfo.moduleMtime',
                       mock_moduleMtime):
                # Ensure first if-block in execute_now returns False
                panoptes_plugin_info.last_results = int(_TIMESTAMP)
                panoptes_plugin_info.last_executed = int(_TIMESTAMP)
                self.assertFalse(panoptes_plugin_info.execute_now)

                # Ensure second if-block in execute_now returns False
                panoptes_plugin_info._last_results = None
                panoptes_plugin_info._last_executed = None

                panoptes_plugin_info.last_results = int(_TIMESTAMP)
                panoptes_plugin_info.last_executed = (int(_TIMESTAMP) - panoptes_plugin_info.execute_frequency)
                self.assertFalse(panoptes_plugin_info.execute_now)

                # Ensure returns True
                panoptes_plugin_info._last_results = None
                panoptes_plugin_info._last_executed = None

                panoptes_plugin_info.last_results = (int(_TIMESTAMP) - panoptes_plugin_info.execute_frequency)
                self.assertTrue(panoptes_plugin_info.execute_now)

    def test_plugin_info_lock(self):
        panoptes_plugin_info = PanoptesPluginInfo("plugin_name", "path/to/plugin")
        panoptes_plugin_info.panoptes_context = self._panoptes_context
        self.assertIsNotNone(panoptes_plugin_info.lock)
        self.assertTrue(panoptes_plugin_info.lock.locked)

        #  Assert lock is cached
        self.assertIsNotNone(panoptes_plugin_info.lock)
        self.assertTrue(panoptes_plugin_info.lock.locked)

        panoptes_plugin_info_2 = PanoptesPluginInfo("plugin_name", "path/to/plugin")
        panoptes_plugin_info_2.panoptes_context = self._panoptes_context

        #  Patch timeout to speed up test
        with patch('yahoo_panoptes.framework.plugins.panoptes_base_plugin.const.PLUGIN_AGENT_LOCK_ACQUIRE_TIMEOUT', 1):
            self.assertFalse(panoptes_plugin_info_2.lock.locked)



"""
Copyright 2018, Oath Inc.
Licensed under the terms of the Apache 2.0 license. See LICENSE file in project root for terms.
"""

from mock import patch, Mock
import unittest
import subprocess
import json  # why isn't the call below recognized?

from yahoo_panoptes.framework.utilities.ping import *

# For testing locally on OSX
TEST_PING_RESPONSE = "ping statistics ---\n" \
                     "3 packets transmitted, 3 received, 0% packet loss, time 1439ms\n" \
                     "rtt min/avg/max/mdev = 0.040/0.120/0.162/0.057 ms"


class TestPing(unittest.TestCase):
    def test_code_output(self):
        # Mock the system calls
        subprocess.check_output = Mock(return_value=TEST_PING_RESPONSE)
        test_ping = PanoptesPing(count=3, timeout=100, hostname="localhost")
        test_d = {u"round_trip_stddev": 0.057, u"packets_received": 3, u"execution_time": 1.44,
                  u"round_trip_avg": 0.120, u"packets_transmitted": 3, u"packet_loss_pct": 0.0,
                  u"round_trip_max": 0.162, u"round_trip_min": 0.040}

        self.assertDictEqual(test_d, json.loads(test_ping.response))
        self.assertEqual(test_ping.packets_transmitted, 3)
        self.assertEqual(test_ping.packets_received, 3)
        self.assertEqual(test_ping.packet_loss_pct, 0.0)
        self.assertEqual(test_ping.execution_time, 1.44)  # add for testing prod
        self.assertEqual(test_ping.round_trip_min, 0.040)
        self.assertEqual(test_ping.round_trip_avg, 0.120)
        self.assertEqual(test_ping.round_trip_max, 0.162)
        self.assertEqual(test_ping.round_trip_stddev, 0.057)

        # Test bad parameters
        with self.assertRaises(AssertionError):
            PanoptesPing(count=-1)
        with self.assertRaises(AssertionError):
            PanoptesPing(count="1")
        with self.assertRaises(AssertionError):
            PanoptesPing(hostname=None)
        with self.assertRaises(AssertionError):
            PanoptesPing(hostname='')
        with self.assertRaises(AssertionError):
            PanoptesPing(timeout=0)
        with self.assertRaises(AssertionError):
            PanoptesPing(timeout=-1)

    def test_ping_timeout(self):
        e = subprocess.CalledProcessError(returncode=None, cmd=None, output=TEST_PING_RESPONSE)
        #  When I try: "TypeError: exceptions must be old-style classes or derived from BaseException, not Mock"
        subprocess.check_output = Mock(side_effect=e)
        test_d = {u"round_trip_stddev": 0.057, u"packets_received": 3, u"execution_time": 1.44,
                  u"round_trip_avg": 0.120, u"packets_transmitted": 3, u"packet_loss_pct": 0.0,
                  u"round_trip_max": 0.162, u"round_trip_min": 0.040}

        with self.assertRaises(PanoptesPingTimeoutException):
            test_timeout_ping = PanoptesPing(count=3, timeout=100, hostname="localhost")
            self.assertDictEqual(test_d, json.loads(test_timeout_ping.response))
            self.assertEqual(test_timeout_ping.packets_transmitted, 3)
            self.assertEqual(test_timeout_ping.packets_received, 3)
            self.assertEqual(test_timeout_ping.execution_time, 1.44)
            self.assertEqual(test_timeout_ping.round_trip_min, 0.040)
            self.assertEqual(test_timeout_ping.round_trip_avg, 0.120)
            self.assertEqual(test_timeout_ping.round_trip_max, 0.162)
            self.assertEqual(test_timeout_ping.round_trip_stddev, 0.057)

    def test_ping_unknown_host(self):
        hostname = 'localhostx'
        unknown_host_response = "ping: cannot resolve " + hostname + ": Unknown host"
        e = subprocess.CalledProcessError(returncode=None, cmd=None,
                                          output=unknown_host_response)  # Shouldn't I be mocking this?
        #  When I try: "TypeError: exceptions must be old-style classes or derived from BaseException, not Mock"
        subprocess.check_output = Mock(side_effect=e)
        test_fail_d = {u"round_trip_stddev": None, u"packets_received": None, u"execution_time": None,
                       u"round_trip_avg": None, u"packets_transmitted": None, u"packet_loss_pct": None,
                       u"round_trip_max": None, u"round_trip_min": None}

        with self.assertRaises(PanoptesPingException):
            test_unknown_host_ping = PanoptesPing(hostname=hostname)
            self.assertDictEqual(test_fail_d, test_unknown_host_ping.response)



"""
Copyright 2018, Oath Inc.
Licensed under the terms of the Apache 2.0 license. See LICENSE file in project root for terms.

This module implements the Panoptes Plugin Manager - a class that can locate and load the plugins specified by plugin
type and plugin class
"""
import imp
import sys

from yapsy.PluginManager import PluginManager

from .. import const
from .panoptes_base_plugin import PanoptesPluginInfo


class PanoptesPluginManager(PluginManager):
    """
    This class implements a Plugin Manager that locates and loads all plugins of the given type

    This extends yapsy's PluginManager by filling in default values and setting PanoptesPluginInfo as the
    plugin_info_class. Also, it attaches the Panoptes Context, Key/Value Store class and any data to the
    PanoptesPluginInfo object so that these are available to the plugin when it is being executed

    Args:
        plugin_type(str): The 'type' of the plugin
        plugin_class(PanoptesBasePlugin): The class of the plugin - should be a subclass of PanoptesBasePlugin
        plugin_info_class(class): The plugin info class - defaults to PanoptesPluginInfo
        plugin_data(object): An optional, arbitrary data object to be passed to the plugin during execution
        panoptes_context(PanoptesContext): Th Panoptes Context of the Plugin Agent executing the plugin
        kv_store_class(class): The Key/Value Store class that the Plugin Info class should use
        secrets_store_class(class): The secrets store class that the Plugin Info class should use
    """

    def __init__(self, plugin_type=None, plugin_class=None, plugin_info_class=PanoptesPluginInfo, plugin_data=None,
                 panoptes_context=None, kv_store_class=None, secrets_store_class=None):
        self._plugin_type = plugin_type
        self._panoptes_context = panoptes_context
        self._kv_store_class = kv_store_class
        self._secrets_store_class = secrets_store_class
        self._data = plugin_data
        self._categories_filter = {plugin_type: plugin_class}
        self._directories_list = panoptes_context.config_dict[plugin_type]['plugins_paths']
        self._panoptes_context.logger.debug(
            'Starting Plugin Manager for "%s" plugins with the following configuration: %s, %s, %s' % (
                plugin_type, self._categories_filter, self._directories_list, const.PLUGIN_EXTENSION))
        super(PanoptesPluginManager, self).__init__(categories_filter=self._categories_filter,
                                                    directories_list=self._directories_list,
                                                    plugin_info_ext=const.PLUGIN_EXTENSION,
                                                    plugin_locator=None)
        self.setPluginInfoClass(plugin_info_class)
        self.collectPlugins()

    def collectPlugins(self):
        """
        This method locates and loads plugins

        Also, it attaches the Panoptes Context, Key/Value Store class and any data to the
        PanoptesPluginInfo object so that these are available to the plugin when it is being executed

        Returns:
            None
        """
        logger = self._panoptes_context.logger
        self.locatePlugins()

        # Set the name of the configuration file so that we can find the mtime for that file while deciding if the
        # the plugin should be executed
        for candidate_infofile, candidate_filepath, plugin_info in self._candidates:
            plugin_info.config_filename = candidate_infofile

        imp.acquire_lock()
        plugins = self.loadPlugins()
        imp.release_lock()
        logger.debug('Found %d plugins' % len(plugins))
        for plugin in plugins:
            logger.debug(
                'Loaded plugin "%s", version "%s" of type "%s", category "%s"' % (
                    plugin.name, plugin.version, self._plugin_type, plugin.category))
            plugin.panoptes_context = self._panoptes_context
            plugin.kv_store_class = self._kv_store_class
            plugin.data = self._data

    def getPluginsOfCategory(self, category_name):
        """
        Plugin Managers compares the current list of plugins returned by Yapsy's PluginManager to the previous list
        and reschedules if the two don't match.  To guarantee that the list order is invariant in order to
        prevent unnecessary re-scheduling, we sort plugins returned by PanoptesPluginManager.getPluginsOfCategory
        by signature.

        Args:
            category_name (str): Plugin category

        Returns:
            list: Sorted list of plugins
        """
        plugins_unfiltered = super(PanoptesPluginManager, self).getPluginsOfCategory(category_name)
        return sorted(plugins_unfiltered, key=lambda obj: obj.signature)

    def __del__(self):
        logger = self._panoptes_context.logger
        for plugin in self.getAllPlugins():
            plugin_module = plugin.plugin_object.__module__
            logger.debug('Deleting module: %s' % plugin_module)
            imp.acquire_lock()
            del sys.modules[plugin_module]
            imp.release_lock()
            del plugin.plugin_object



from yahoo_panoptes.enrichment.enrichment_plugin import PanoptesEnrichmentPlugin
from yahoo_panoptes.enrichment.schema.interface import PanoptesInterfaceEnrichmentGroup
from yahoo_panoptes.framework.enrichment import PanoptesEnrichmentSet, PanoptesEnrichmentGroupSet
from yahoo_panoptes.framework.plugins.base_snmp_plugin import PanoptesSNMPBaseEnrichmentPlugin
from yahoo_panoptes.framework.utilities.helpers import transform_octet_to_mac
from yahoo_panoptes.framework.utilities.snmp.mibs.ifTable import getIfTypeDesc, ifDescr, ifType, ifSpeed, ifPhysAddress
from yahoo_panoptes.framework.utilities.snmp.mibs.ifXTable import ifAlias, ifName, ifHighSpeed


class PluginEnrichmentInterface(PanoptesSNMPBaseEnrichmentPlugin, PanoptesEnrichmentPlugin):
    _MISSING_VALUE_STRING = u'<not set>'
    _MISSING_METRIC_VALUE = -1
    interface_enrichment_oids = [ifType, ifDescr, ifName, ifAlias, ifHighSpeed, ifSpeed, ifPhysAddress]

    def __init__(self):
        super(PluginEnrichmentInterface, self).__init__()
        self._interface_enrichment_group = None
        self._interface_enrichment_group_set = None
        self._enrichments_map = dict()
        self._interface_table = dict()

    def _build_enrichments_map(self):
        self._enrichments_map = dict()
        for enrichment_oid in self.interface_enrichment_oids:
            for varbind in self._snmp_connection.bulk_walk(enrichment_oid,
                                                           max_repetitions=self.snmp_configuration.max_repetitions):
                self._enrichments_map[varbind.oid + '.' + varbind.index] = varbind.value

    def get_interface_name(self, index):
        return self._enrichments_map.get(ifName + '.' + index, self._MISSING_VALUE_STRING)

    @property
    def host(self):
        return self._host

    def get_parent_interface_name(self, index):
        """
        Gets the parent interface name for the interface associated with the provided index

        Args:
            index (int): The index used to look up the associated interface in self._interface_table

        Returns:
            string: The name of the parent interface, or self._MISSING_VALUE_STRING if the interface has no parent.
                    For Cisco devices, this is everything to the left of the '.' in the interface name, if a '.' is
                    present.
        """
        return self._MISSING_VALUE_STRING

    def get_parent_interface_media_type(self, index):
        """
        Gets the parent interface media type for the interface associated with the provided index

        Args:
            index (int): The index used to look up the associated interface in self._interface_table

        Returns:
            string: The media_type of the parent interface, or self._MISSING_VALUE_STRING if the interface
            has no parent.
        """
        parent_index = self._get_parent_interface_index(index)
        if parent_index:
            return self._interface_table[parent_index]['media_type']
        else:
            return self._MISSING_VALUE_STRING

    def get_parent_interface_port_speed(self, index):
        """
        Gets the parent interface port speed for the interface associated with the provided index

        Args:
            index (int): The index used to look up the associated interface in self._interface_table

        Returns:
            integer: The port speed of the parent interface, or self._MISSING_METRIC_VALUE if the interface
            has no parent.
        """
        parent_index = self._get_parent_interface_index(index)
        if parent_index:
            return self._interface_table[parent_index]['port_speed']
        else:
            return self._MISSING_METRIC_VALUE

    def get_parent_interface_configured_speed(self, index):
        """
        Gets the parent interface configured speed for the interface associated with the provided index

        Args:
            index (int): The index used to look up the associated interface in self._interface_table

        Returns:
            integer: The configured speed of the parent interface, or self._MISSING_METRIC_VALUE if the interface
            has no parent.
        """
        parent_index = self._get_parent_interface_index(index)
        if parent_index:
            return self._interface_table[parent_index]['configured_speed']
        else:
            return self._MISSING_METRIC_VALUE

    def get_description(self, index):
        return self._enrichments_map.get(ifDescr + '.' + index, self._MISSING_VALUE_STRING)

    def get_media_type(self, index):
        type_index = self._enrichments_map.get(ifType + '.' + index)
        return getIfTypeDesc(type_index) if type_index is not None else self._MISSING_VALUE_STRING

    def get_alias(self, index):
        alias = self._enrichments_map.get(ifAlias + '.' + index, self._MISSING_VALUE_STRING)
        return alias if len(alias) > 0 else self._MISSING_VALUE_STRING

    def get_if_speed(self, index):
        return int(self._enrichments_map.get(ifSpeed + '.' + index, self._MISSING_METRIC_VALUE))

    def get_configured_speed(self, index):
        high_speed = self._enrichments_map.get(ifHighSpeed + '.' + index)
        if high_speed in ['0', '1', None]:
            speed = self.get_if_speed(index)
            return int(speed) if speed != 0 else (int(high_speed) * 1000000 if high_speed is not None
                                                  else self._MISSING_METRIC_VALUE)
        else:
            return int(high_speed) * 1000000  # Mbps to bps

    def get_port_speed(self, index):
        return self.get_configured_speed(index)

    def get_physical_address(self, index):
        physical_address = self._enrichments_map.get(ifPhysAddress + '.' + index, self._MISSING_VALUE_STRING)
        return transform_octet_to_mac(physical_address) if physical_address not in [None, ""] else \
            self._MISSING_VALUE_STRING

    @property
    def interface_table(self):
        return self._interface_table

    def _build_interface_table(self):
        for oid in self._enrichments_map.keys():
            index = oid.split('.')[-1]
            self._interface_table[index] = dict()
            self._interface_table[index]['interface_name'] = self.get_interface_name(index)
            self._interface_table[index]['description'] = self.get_description(index)
            self._interface_table[index]['media_type'] = self.get_media_type(index)
            self._interface_table[index]['alias'] = self.get_alias(index)
            self._interface_table[index]['configured_speed'] = self.get_configured_speed(index)
            self._interface_table[index]['port_speed'] = self.get_port_speed(index)
            self._interface_table[index]['physical_address'] = self.get_physical_address(index)

    def _get_parent_interface_index(self, index):
        parent_interface_name = self.get_parent_interface_name(index)
        return self._get_index_from_interface_name(parent_interface_name)

    def _get_index_from_interface_name(self, name):
        for index in self._interface_table.keys():
            if name == self._interface_table[index]['interface_name']:
                return index
        return None

    def _add_parent_interface_enrichments(self):
        for oid in self._interface_table.keys():
            index = oid.split('.')[-1]
            self._interface_table[index]['parent_interface_name'] = self.get_parent_interface_name(index)
            self._interface_table[index]['parent_interface_media_type'] = self.get_parent_interface_media_type(index)
            self._interface_table[index]['parent_interface_configured_speed'] = \
                self.get_parent_interface_configured_speed(index)
            self._interface_table[index]['parent_interface_port_speed'] = self.get_parent_interface_port_speed(index)

    def get_results(self):
        self._interface_enrichment_group = PanoptesInterfaceEnrichmentGroup(enrichment_ttl=self.enrichment_ttl,
                                                                            execute_frequency=self.execute_frequency)
        self._interface_enrichment_group_set = PanoptesEnrichmentGroupSet(self.resource)

        self._build_enrichments_map()
        self._build_interface_table()
        self._add_parent_interface_enrichments()

        for index, enrichment_set in self.interface_table.items():
            try:
                self._interface_enrichment_group.add_enrichment_set(PanoptesEnrichmentSet(str(index), enrichment_set))
            except Exception as e:
                self._logger.error('Error while adding enrichment set {} to enrichment group for the device {}: {}'.
                                   format(str(index), self.host, repr(e)))

        self._interface_enrichment_group_set.add_enrichment_group(self._interface_enrichment_group)

        self._logger.debug('Interface enrichment for device {} PanoptesEnrichmentGroupSet {}'.
                           format(self.host, self._interface_enrichment_group_set))

        return self._interface_enrichment_group_set



"""
Copyright 2018, Oath Inc.
Licensed under the terms of the Apache 2.0 license. See LICENSE file in project root for terms.

This module defines Panoptes enrichment and their related abstractions
"""

import copy
import json
import time

from cerberus import Validator

from . import const
from .context import PanoptesContext
from .exceptions import PanoptesBaseException
from .resources import PanoptesResource
from .utilities.key_value_store import PanoptesKeyValueStore
from .validators import PanoptesValidators


class PanoptesEnrichmentException(PanoptesBaseException):
    """
    The base class for all Panoptes Enrichment exceptions
    """
    pass


class PanoptesEnrichmentCacheKeyValueStore(PanoptesKeyValueStore):
    """
    A custom Key/Value store for Panoptes Enrichment Cache
    """
    redis_group = const.ENRICHMENT_REDIS_GROUP

    def __init__(self, panoptes_context):
        super(PanoptesEnrichmentCacheKeyValueStore, self).__init__(
            panoptes_context, const.ENRICHMENT_PLUGIN_RESULTS_KEY_VALUE_NAMESPACE)


class PanoptesEnrichmentSchemaValidator(object):
    """
    Schema validator base class based on Cerberus
    """
    schema = {'key': 'value'}

    def __init__(self):
        assert isinstance(self.schema, dict) and len(self.schema) > 0, 'schema must be a non empty Cerberus schema dict'
        self.__cerberus_validator = Validator(schema=self.schema)

    def validate(self, enrichment_set_data):
        """
        Validates PanoptesEnrichmentSet object against defined schema

        Args:
            enrichment_set_data (PanoptesEnrichmentSet): PanoptesEnrichmentSet to validate
        Returns:
              bool
        """
        assert isinstance(enrichment_set_data, PanoptesEnrichmentSet), \
            'element set must be an instance of PanoptesEnrichmentSet'
        schema = copy.deepcopy(self.schema)
        schema[enrichment_set_data.key] = schema.pop('enrichment_label')
        return self.__cerberus_validator.validate(document=enrichment_set_data._raw_data, schema=schema)


class PanoptesEnrichmentEncoder(json.JSONEncoder):
    """
    Custom Json encoder to convert set to list during encoding
    """
    # https://github.com/PyCQA/pylint/issues/414
    def default(self, o):  # pylint: disable=E0202
        if isinstance(o, set):
            return list(o)
        if isinstance(o, PanoptesResource):
            return o.__dict__['_PanoptesResource__data']
        if isinstance(o, PanoptesEnrichmentSet):
            return o.__dict__['_PanoptesEnrichmentSet__data']
        if isinstance(o, PanoptesEnrichmentGroup):
            return o.__dict__['_PanoptesEnrichmentGroup__data']
        if isinstance(o, PanoptesEnrichmentGroupSet):
            return o.__dict__['_PanoptesEnrichmentGroupSet__data']
        if isinstance(o, PanoptesEnrichmentMultiGroupSet):
            return o.__dict__['_PanoptesEnrichmentMultiGroupSet__data']
        return json.JSONEncoder.default(self, o)


class PanoptesEnrichmentSet(object):
    """
    Representation of a basic enrichment element

    A PanoptesEnrichmentSet hold key(label) and collection of enrichment info as its value

    Args:
        key (as defined in schema): label for enrichment element collection
        value (dict): enrichment collection

    Examples:
        An example enrichment set data structure

    {
       "int_001" : {
          "speed" : 1000,
          "index" : 1,
          "status" : "up"
       }
    }
    """
    def __init__(self, key, value={None: None}):
        assert PanoptesValidators().valid_nonempty_string(key), 'enrichment key must be a string'
        assert isinstance(value, dict), 'enrichment value must be a dict'
        if value == {None: None}:
            value = {}
        self.__data = dict()
        self._key = key
        self.__data[self._key] = value

    @property
    def key(self):
        return self._key

    @property
    def value(self):
        return self.__data[self._key]

    @property
    def _raw_data(self):
        return self.__data

    def add(self, enrichment_key, enrichment_value):
        """
        Adds enrichment key and value elements to PanoptesEnrichmentSet object

        Args:
            enrichment_key (as defined in schema): The enrichment element key
            enrichment_value (as defined in schema): The enrichment element value corresponding to the key

        Returns:
            None

        """
        self.__data[self._key][enrichment_key] = enrichment_value

    def json(self):
        return json.dumps(self.__data, sort_keys=True)

    def __repr__(self):
        return self.__class__.__name__ + '({0})'.format(str(self.__data))

    def __hash__(self):
        return hash(self._key)

    def __len__(self):
        return len(self.__data[self._key])

    def __eq__(self, other):
        if not isinstance(other, PanoptesEnrichmentSet):
            return False
        return self._key == other._key


class PanoptesEnrichmentGroup(object):
    """
    Representation of Enrichment Group

    Collection of enrichment elements(PanoptesEnrichmentSet) grouped by Enrichment key(namespace)

    Args:
        namespace (str): enrichment namespace
        schema_validator (PanoptesEnrichmentSchemaValidator): PanoptesEnrichmentSchemaValidator
        instance initialized with schema
        execute_frequency (int): Execute frequency of the plugin which produced this object
        enrichment_ttl (int): TTL value of the enrichment object

    Examples:
        An example PanoptesEnrichmentGroup data structure

    {
       "namespace" : "interface",
       "data" : [
          {
             "value" : {
                "speed" : 1000,
                "status" : "up",
                "index" : 1
             },
             "key" : "int_001"
          }
       ]
    }
    """

    def __init__(self, namespace, schema_validator, enrichment_ttl, execute_frequency):
        assert PanoptesValidators().valid_nonempty_string(namespace), 'enrichment namespace must be a string'
        assert isinstance(schema_validator, PanoptesEnrichmentSchemaValidator), \
            'schema_validator must be an instance of PanoptesEnrichmentSchemaValidator'
        assert PanoptesValidators().valid_nonzero_integer(enrichment_ttl), \
            'enrichment_ttl must be a valid nonzero integer'
        assert PanoptesValidators().valid_nonzero_integer(execute_frequency), \
            'execute_frequency must be a valid nonzero integer'
        self.__data = dict()
        self.__data['metadata'] = dict()
        self.__data['namespace'] = namespace
        self.__data['data'] = set()
        self.__schema_validator = schema_validator
        self.__data['metadata']['_enrichment_group_creation_timestamp'] = time.time()
        self.__data['metadata']['_enrichment_ttl'] = enrichment_ttl
        self.__data['metadata']['_execute_frequency'] = execute_frequency

    @property
    def enrichment_schema(self):
        return self.__schema_validator.schema

    @property
    def namespace(self):
        return self.__data['namespace']

    @property
    def data(self):
        return self.__data['data']

    @property
    def _raw_data(self):
        return self.__data

    @property
    def metadata(self):
        return self.__data['metadata']

    @property
    def enrichment_ttl(self):
        return self.__data['metadata']['_enrichment_ttl']

    @property
    def execute_frequency(self):
        return self.__data['metadata']['_execute_frequency']

    @property
    def enrichment_group_creation_timestamp(self):
        return self.__data['metadata']['_enrichment_group_creation_timestamp']

    def add_enrichment_set(self, enrichment_set):
        """
        Adds enrichment_set(PanoptesEnrichmentSet) to create Enrichment Group

        Args:
            enrichment_set (PanoptesEnrichmentSet): Collection of enrichment elements(PanoptesEnrichmentSet)

        Returns:
            None
        """
        assert isinstance(enrichment_set, PanoptesEnrichmentSet), \
            'enrichment set must be an instance of PanoptesEnrichmentSet'

        assert self.__schema_validator.validate(enrichment_set), \
            'schema validation failed for enrichment_set data'
        self.__data['data'].discard(enrichment_set)
        self.__data['data'].add(enrichment_set)

    def upsert_metadata(self, metadata_key, metadata_value):
        """
        Adds metadata key / value pairs to PanoptesEnrichmentGroup

        Args:
            metadata_key (str): metadata key
            metadata_value (str or int or float):  metadata value

        Returns:
              None
        """
        assert PanoptesValidators().valid_nonempty_string(metadata_key), 'metadata_key must be a string'

        if metadata_key.startswith('_'):
            raise ValueError('Failed to update reserved metadata')

        assert \
            PanoptesValidators().valid_nonempty_string(metadata_value) or \
            PanoptesValidators().valid_number(metadata_value), 'metadata_value must be any of string / float / integer'

        self.__data['metadata'][metadata_key] = metadata_value

    def bulk_add_enrichment_set(self):
        pass

    def json(self):
        return json.dumps(self.__data, sort_keys=True, cls=PanoptesEnrichmentEncoder)

    def serialize_data(self):
        return json.dumps({enrichment_set.key: enrichment_set.value for enrichment_set in self.data}, sort_keys=True)

    def serialize(self):
        enrichment_serialize = dict()
        enrichment_serialize['data'] = {enrichment_set.key: enrichment_set.value for enrichment_set in self.data}
        enrichment_serialize['metadata'] = self.__data['metadata']
        return json.dumps(enrichment_serialize, sort_keys=True)

    def __repr__(self):
        return self.__class__.__name__ + '({0})'.format(str(self.__data))

    def __len__(self):
        return len(self.__data['data'])

    def __hash__(self):
        return hash(self.namespace)

    def __eq__(self, other):
        if not isinstance(other, PanoptesEnrichmentGroup):
            return False
        return self.namespace == other.namespace


class PanoptesEnrichmentGroupSet(object):
    """
    Representation of Enrichment Group Set

    Collection of enrichment group grouped by resource

    Args:
        resource (PanoptesResource): PanoptesResource object

    Example:
        An example PanoptesEnrichmentGroupSet data structure
    {
       "resource" : PanoptesResource_instance,
       "enrichment" : [
          {
             "namespace" : "interface",
             "data" : [
                {
                   "value" : {
                      "index" : 1,
                      "status" : "up",
                      "speed" : 1000
                   },
                   "key" : "int_001"
                }
             ]
          }
       ]
    }
    """

    def __init__(self, resource):
        assert isinstance(resource, PanoptesResource), 'resource must be an instance of PanoptesResource'
        self.__data = dict()
        self.__data['resource'] = resource
        self.__data['enrichment_group_set_creation_timestamp'] = time.time()
        self.__data['enrichment'] = set()

    @property
    def resource(self):
        return self.__data['resource']

    @property
    def enrichment(self):
        return self.__data['enrichment']

    @property
    def enrichment_group_set_creation_timestamp(self):
        return self.__data['enrichment_group_set_creation_timestamp']

    @property
    def _raw_data(self):
        return self.__data

    def add_enrichment_group(self, enrichment_group):
        """
        Adds enrichment_group(PanoptesEnrichmentGroup) to create enrichment group set

        Args:
            enrichment_group (PanoptesEnrichmentGroup): Enrichment elements grouped by key

        Returns:
            None
        """
        assert isinstance(enrichment_group, PanoptesEnrichmentGroup), \
            'enrichment_group must be an instance of PanoptesEnrichmentGroup'
        assert len(enrichment_group.data) > 0, 'enrichment_group must hold at least one data set'
        self.__data['enrichment'].discard(enrichment_group)
        self.__data['enrichment'].add(enrichment_group)

    def bulk_add_enrichment_group(self):
        pass

    def json(self):
        return json.dumps(self.__data, sort_keys=True, cls=PanoptesEnrichmentEncoder)

    def __repr__(self):
        return self.__class__.__name__ + '({0})'.format(str(self.__data))

    def __len__(self):
        return len(self.__data['enrichment'])

    def __hash__(self):
        namespaces_self = ''.join(sorted([item.namespace for item in self.enrichment]))
        return hash(self.resource.resource_id + namespaces_self)

    def __eq__(self, other):
        if not isinstance(other, PanoptesEnrichmentGroupSet):
            return False
        namespaces_other = ''.join(sorted([item.namespace for item in other.enrichment]))
        namespaces_self = ''.join(sorted([item.namespace for item in self.enrichment]))
        return self.resource.resource_id == other.resource.resource_id and namespaces_self == namespaces_other


class PanoptesEnrichmentMultiGroupSet(object):
    """
    Collection of PanoptesEnrichmentGroupSet belongs to multiple Panoptes resources
    """
    def __init__(self):
        self.__data = dict()
        self.__data['group_sets'] = set()

    def add_enrichment_group_set(self, enrichment_group_set):
        """
        Adds enrichment_group_set(PanoptesEnrichmentGroupSet)

        Args:
            enrichment_group_set (PanoptesEnrichmentGroupSet): Enrichment group set of a Panoptes resource

        Returns:
            None
        """
        assert isinstance(enrichment_group_set, PanoptesEnrichmentGroupSet), \
            'enrichment_group_set must be an instance of PanoptesEnrichmentGroupSet'
        assert len(enrichment_group_set) > 0, 'enrichment_group_set must hold at least one data set'
        self.__data['group_sets'].discard(enrichment_group_set)
        self.__data['group_sets'].add(enrichment_group_set)

    @property
    def enrichment_group_sets(self):
        return self.__data['group_sets']

    def __len__(self):
        return len(self.__data['group_sets'])

    def __repr__(self):
        return self.__class__.__name__ + '({0})'.format(str(self.__data))

    def json(self):
        return json.dumps(self.__data, sort_keys=True, cls=PanoptesEnrichmentEncoder)


class PanoptesEnrichmentCacheError(PanoptesBaseException):
    """
    The base class for all Panoptes Enrichment Cache errors
    """
    pass


class PanoptesEnrichmentCache(object):
    """
    Fetches enrichment from kv store and act as cache for polling plugins

    Args:
        panoptes_context (PanoptesContext): The PanoptesContext being used by the Plugin Agent
        plugin_conf (dict): Plugin conf dict
        resource (PanoptesResource): Resource object associated with plugin runner instance
    """

    def __init__(self, panoptes_context, plugin_conf, resource):
        assert isinstance(panoptes_context, PanoptesContext), 'panoptes_context must be an instance of PanoptesContext'
        assert isinstance(resource, PanoptesResource), 'resource must be an instance of PanoptesResource'
        assert isinstance(plugin_conf, dict), 'plugin_conf value must be a dict'

        self._logger = panoptes_context.logger
        self._plugin_name = plugin_conf['Core']['name']
        self._panoptes_context = panoptes_context
        self._enrichment_conf = plugin_conf['enrichment']
        self._resource_id = resource.resource_id
        self._preload_conf = self._parse_conf()
        self._enrichment_data = dict()

        try:
            self._kv_store = PanoptesEnrichmentCacheKeyValueStore(panoptes_context)
        except Exception as e:
            raise e

        self._process_enrichment()

        self._logger.debug('Successfully created PanoptesEnrichmentCache enrichment_data {} for plugin {}'.
                           format(self._enrichment_data, self._plugin_name))

    def get_enrichment(self, resource_id, namespace):
        """
        Returns enrichment data for matching resource_id and namespace

        Args:
            resource_id (str): Panoptes resource id
            namespace (str): Enrichment namespace
        Returns:
              enrichment_data (dict): enrichment_data dict
        """
        assert PanoptesValidators().valid_nonempty_string(resource_id), 'resource_id must be a string'
        assert PanoptesValidators().valid_nonempty_string(namespace), 'enrichment namespace must be a string'

        if resource_id == 'self':
            resource_id = self._resource_id

        try:
            return {key: value for key, value in self._enrichment_data[resource_id][namespace].items()}
        except KeyError:
            try:
                self._preload_data(resource_id, namespace)
                return {key: value for key, value in self._enrichment_data[resource_id][namespace].items()}
            except Exception as e:
                raise PanoptesEnrichmentCacheError('Failed to get data for resource {} namespace {} from enrichment '
                                                   'resource object: {}'.format(resource_id, namespace, repr(e)))

    def get_enrichment_value(self, resource_id, namespace, enrichment_key):
        """
        Returns enrichment value for matching resource_id, namespace and
        enrichment_key

        Args:
            resource_id (str): Panoptes resource id
            namespace (str): Enrichment namespace
            enrichment_key (str): Enrichment set key
        Returns:
              enrichment_set (dict): key, value pairs of enrichment
        """
        assert PanoptesValidators().valid_nonempty_string(resource_id), 'resource_id must be a string'
        assert PanoptesValidators().valid_nonempty_string(namespace), 'enrichment namespace must be a string'
        assert PanoptesValidators().valid_nonempty_string(enrichment_key), 'enrichment_key must be a string'

        if resource_id == 'self':
            resource_id = self._resource_id

        try:
            return self._enrichment_data[resource_id][namespace][enrichment_key]
        except KeyError:
            try:
                self._preload_data(resource_id, namespace)
                return self._enrichment_data[resource_id][namespace][enrichment_key]
            except Exception as e:
                raise PanoptesEnrichmentCacheError(
                    'Failed to get data for resource {} namespace {} enrichment_key {} '
                    'from enrichment cache object: {}'.format(resource_id, namespace, enrichment_key, repr(e)))

    def get_enrichment_keys(self, resource_id, namespace):
        """
        Returns enrichment keys for matching resource_id and namespace

        Args:
            resource_id (str): Panoptes resource id
            namespace (str): Enrichment namespace
        Returns:
              enrichment_keys (list): enrichment keys(label)
        """
        assert PanoptesValidators().valid_nonempty_string(resource_id), 'resource_id must be a string'
        assert PanoptesValidators().valid_nonempty_string(namespace), 'enrichment namespace must be a string'

        if resource_id == 'self':
            resource_id = self._resource_id

        try:
            return self._enrichment_data[resource_id][namespace].keys()
        except KeyError:
            try:
                self._preload_data(resource_id, namespace)
                return self._enrichment_data[resource_id][namespace].keys()
            except Exception as e:
                raise PanoptesEnrichmentCacheError('Failed to get data for resource {} namespace {} from enrichment '
                                                   'cache object: {}'.format(resource_id, namespace, repr(e)))

    def _process_enrichment(self):
        for resource, namespace in self._preload_conf:
            if resource == 'self':
                resource = self._resource_id

            namespace_keys = None
            if namespace == '*':
                try:
                    namespace_keys = self._kv_store.find_keys(resource + const.KV_NAMESPACE_DELIMITER + namespace)
                except Exception as e:
                    self._logger.error(
                        'Error while scanning namespace pattern {} on KV store for plugin {} resource {}: {}'.format(
                            namespace, self._plugin_name, resource, repr(e)))
                if namespace_keys:
                    namespaces = [namespace_field.split(':')[-1] for namespace_field in namespace_keys]
                    for normalized_namespace in namespaces:
                        self._preload_data(resource, normalized_namespace)
            else:
                self._preload_data(resource, namespace)

    def _preload_data(self, resource, namespace):
        try:
            key = resource + const.KV_NAMESPACE_DELIMITER + namespace
            value = self._kv_store.get(key)
            if value:
                data = json.loads(value).get('data')
                self._enrichment_data.setdefault(resource, dict())
                self._enrichment_data[resource].update(**{namespace: data})
                self._logger.debug('Successfully populated enrichment for plugin {} resource {} namespace {} data {}'.
                                   format(self._plugin_name, resource, namespace, data))
            else:
                self._logger.error(
                    'No enrichment data found on KV store for plugin {} resource {} namespace {} using key {}'.format(
                        self._plugin_name, resource, namespace, key))
        except Exception as e:
            raise IOError(
                'Failed while pre-loading enrichment from KV store for plugin {} resource {} namespace {}: {}'.format(
                    self._plugin_name, resource, namespace, repr(e)))

    def _parse_conf(self):
        try:
            return {(item.split(':')[0].strip(), item.split(':')[1].strip())
                    for item in self._enrichment_conf.get('preload').split(',')}
        except Exception as e:
            raise ('Failed while parsing preload enrichment configuration from plugin conf for '
                   'plugin {}: {}'.format(self._plugin_name, repr(e)))



"""
Copyright 2018, Oath Inc.
Licensed under the terms of the Apache 2.0 license. See LICENSE file in project root for terms.

This module implements the Discovery Plugin Agent which accepts plugin names as Celery Task parameters and executes
them. Results, if any, returned by each plugin are expected to be PanoptesResourceSets. If the returned Resource Set is
not empty, each resource from the resource set is placed on a Kafka queue named '<resource_site>-discovery'

This module is expected to be imported and executed though the Celery 'worker' command line tool
"""
import faulthandler
import json
import sys

from celery import task
from celery.signals import worker_shutdown

from ..framework import const
from ..framework.exceptions import PanoptesBaseException
from ..framework.context import PanoptesContext
from ..framework.celery_manager import PanoptesCeleryConfig, PanoptesCeleryInstance
from ..framework.resources import PanoptesResourceSet
from ..framework.plugins.panoptes_base_plugin import PanoptesPluginInfo
from ..framework.plugins.runner import PanoptesPluginWithEnrichmentRunner
from ..framework.utilities.helpers import get_calling_module_name
from ..framework.utilities.key_value_store import PanoptesKeyValueStore
from ..framework.utilities.secrets import PanoptesSecretsStore
from .panoptes_discovery_plugin import PanoptesDiscoveryPlugin

panoptes_context = None
panoptes_discovery_task_context = None
celery = None


class PanoptesDiscoveryPluginAgentKeyValueStore(PanoptesKeyValueStore):
    """
    A custom Key/Value store for the Discovery Plugin Agent which store data like a plugin's last execution time and \
    the last result production time
    """

    def __init__(self, panoptes_context):
        super(PanoptesDiscoveryPluginAgentKeyValueStore, self).\
            __init__(panoptes_context, const.DISCOVERY_PLUGIN_AGENT_KEY_VALUE_NAMESPACE)


class PanoptesDiscoveryPluginKeyValueStore(PanoptesKeyValueStore):
    """
    A custom Key/Value store for the Panoptes Discovery plugins

    Plugins executed by the Discovery Plugin Agent get access to this store and can get/set values they need for their\
    operation
    """

    def __init__(self, panoptes_context):
        super(PanoptesDiscoveryPluginKeyValueStore, self).__init__(panoptes_context, const.PLUGINS_KEY_VALUE_NAMESPACE)


class PanoptesDiscoveryAgentContext(PanoptesContext):
    """
    This class implements a PanoptesContext with no clients - only used for setting up the Celery app
    """

    def __init__(self):
        super(PanoptesDiscoveryAgentContext, self).__init__(key_value_store_class_list=[],
                                                            create_message_producer=False,
                                                            create_zookeeper_client=False)


class PanoptesDiscoveryTaskContext(PanoptesContext):
    """
    This class implements a PanoptesContext with the following:
        - A Key/Value Store for Discovery Plugin Agent
        - A Key/Value Store for Panoptes Plugins
        - A Asynchronous Message Producer
        - A Zookeeper Client
    """

    def __init__(self):
        super(PanoptesDiscoveryTaskContext, self).__init__(
            key_value_store_class_list=[PanoptesDiscoveryPluginAgentKeyValueStore,
                                        PanoptesDiscoveryPluginKeyValueStore,
                                        PanoptesSecretsStore],
            create_message_producer=True, async_message_producer=True, create_zookeeper_client=True)


class PanoptesDiscoveryPluginAgentError(PanoptesBaseException):
    """
    The exception class for Discovery Plugin Agent runtime errors
    """
    pass


@task
def discovery_plugin_task(discovery_plugin_name):
    """
    The main method of the Discovery Agent

    This method, called by Celery, loads and executes the specified plugin. The workflow is as follows:
        - Locate and load plugin through `yapsy's <http://yapsy.sourceforge.net/>`_ plugin manager
        - Evaluate if the plugin should be executed right now
        - If yes, then create a plugin context
        - Attempt to get a lock for the instance of the plugin (unique combination of the plugin name and configuration)
        - If the lock is acquired, then attempt to execute the plugin
        - If the plugin executes successfully, update the timestamp in the Discovery Plugin Agent Key/Value store
        - If the plugin produced a non-zero result set, send the resource over the message bus
        - If sending over the message succeeds, update the timestamp in the Discovery Plugin Agent Key/Value store

    Args:
        discovery_plugin_name (str): The name of the plugin to be executed

    Returns:
        PanoptesResourceSet

    """
    global panoptes_discovery_task_context

    if panoptes_discovery_task_context is None:
        try:
            panoptes_discovery_task_context = PanoptesDiscoveryTaskContext()
        except Exception as e:
            sys.exit('Could not create a Panoptes Discovery Task Context: %s' % (repr(e)))

    logger = panoptes_context.logger

    logger.debug('panoptes_context object: %s' % panoptes_context)

    try:
        plugin_runner = PanoptesPluginWithEnrichmentRunner(
            plugin_name=discovery_plugin_name,
            plugin_type='discovery',
            plugin_class=PanoptesDiscoveryPlugin,
            plugin_info_class=PanoptesPluginInfo,
            plugin_data=None,
            panoptes_context=panoptes_discovery_task_context,
            plugin_agent_kv_store_class=PanoptesDiscoveryPluginAgentKeyValueStore,
            plugin_kv_store_class=PanoptesDiscoveryPluginKeyValueStore,
            plugin_secrets_store_class=PanoptesSecretsStore,
            plugin_logger_name=const.DISCOVERY_PLUGIN_AGENT_PLUGINS_CHILD_LOGGER_NAME,
            plugin_result_class=PanoptesResourceSet,
            results_callback=__send_resource_set)
        plugin_runner.execute_plugin()
    except Exception as e:
        logger.error('[%s] Error executing plugin: %s' % (discovery_plugin_name, str(e)))


def __send_resource_set(context, results, plugin):
    """
    Emits each resource from the provided resource set to message bus

    Args:
        context (PanoptesContext): The PanoptesContext being used by the Plugin Agent
        results (PanoptesResourceSet): The ResourceSet returned by the plugin
        plugin (PanoptesPluginInfo): The plugin object that produced this set of results

    Returns:
        None
    """
    logger = context.logger
    producer = context.message_producer
    key = str(plugin.normalized_name) + ':' + str(plugin.version) + ':' + str(plugin.signature)
    resources_by_site = results.get_resources_by_site()
    logger.debug('Results: %s' % str(resources_by_site))
    logger.info('Found %d sites in results' % len(resources_by_site))
    for resource_site in resources_by_site.keys():
        topic = resource_site + '-' + 'resources'
        resource_list = dict(resources=[resource.raw for resource in resources_by_site[resource_site]])
        resource_list['resource_set_creation_timestamp'] = results.resource_set_creation_timestamp
        resource_list['resource_set_schema_version'] = results.resource_set_schema_version
        logger.debug('Going to send resource "%s" to message bus' % resource_list)
        try:
            producer.send_messages(topic, key, json.dumps(resource_list))
        except:
            raise
        logger.debug('Sent resource "%s" to message bus' % resource_list)


@worker_shutdown.connect()
def shutdown_signal_handler(sender, args=None, **kwargs):
    """
    This method handles the shutdown signal from Celery

    Args:
        sender (Celery): The Celery app which sent the shutdown signal
        args (dict):
        **kwargs (dict):

    Returns:
        None
    """
    global panoptes_context
    panoptes_context.logger.info('Discovery Plugin Agent Worker shutting down - going to stop message producer')
    try:
        panoptes_context.message_producer.stop()
    except Exception as e:
        panoptes_context.logger.error('Could not shutdown message producer: %s' % str(e))


def start_discovery_plugin_agent():
    """
    The entry point for the Discovery Plugin Agent

    This method creates a Panoptes Context and the Celery Instance for the Discovery Plugin Agent

    Returns:
        None
    """
    global panoptes_context, celery
    try:
        panoptes_context = PanoptesDiscoveryAgentContext()
    except Exception as e:
        sys.exit('Could not create a Panoptes Context: %s' % (str(e)))

    logger = panoptes_context.logger
    logger.info('Attempting to start Celery application')

    celery_config = PanoptesCeleryConfig(const.DISCOVERY_PLUGIN_AGENT_CELERY_APP_NAME)

    try:
        celery = PanoptesCeleryInstance(panoptes_context, celery_config).celery
    except Exception as exp:
        sys.exit('Could not instantiate Celery application: %s' % str(exp))
    else:
        logger.info('Started Celery application: %s' % celery)


if get_calling_module_name() == const.CELERY_LOADER_MODULE:
    faulthandler.enable()
    start_discovery_plugin_agent()



"""
Copyright 2018, Oath Inc.
Licensed under the terms of the Apache 2.0 license. See LICENSE file in project root for terms.

The module implements the Discovery Plugin Scheduler which finds any Discovery Plugins in the configured path and then
installs/updates a Celery Beat schedule for each discovery plugin and configuration combination.

If the list or the configuration of the plugins changes, the schedule is updated

This module is expected to be imported and executed though the Celery 'beat' command line tool

Internally, there are two threads that get setup: one is the main thread that runs the Celery Beat service. The other is
the thread started by the Discovery Plugin Scheduler to detect and update plugin/configuration changes
"""
import faulthandler
import sys
import time
from datetime import timedelta

from celery.signals import beat_init

from ..framework import const
from ..framework.context import PanoptesContext
from ..framework.celery_manager import PanoptesCeleryConfig
from ..framework.plugins.helpers import expires, time_limit
from ..framework.plugins.manager import PanoptesPluginManager
from ..framework.plugins.panoptes_base_plugin import PanoptesPluginConfigurationError, \
    PanoptesPluginInfo
from ..framework.plugins.scheduler import PanoptesPluginScheduler
from ..framework.utilities.helpers import get_calling_module_name
from ..framework.utilities.key_value_store import PanoptesKeyValueStore
from .panoptes_discovery_plugin import PanoptesDiscoveryPlugin

panoptes_context = None
discovery_plugin_scheduler = None
celery = None
logger = None


class PanoptesDiscoveryPluginSchedulerKeyValueStore(PanoptesKeyValueStore):
    """
    A custom Key/Value store for the Discovery Plugin Scheduler which uses the namespace demarcated for Discovery
    Plugin Scheduler
    """

    def __init__(self, context):
        super(PanoptesDiscoveryPluginSchedulerKeyValueStore, self).__init__(
                context, const.DISCOVERY_PLUGIN_SCHEDULER_KEY_VALUE_NAMESPACE)


class PanoptesDiscoveryPluginSchedulerContext(PanoptesContext):
    """
    This class implements a PanoptesContext with the following:
        A Key/Value Store for Discovery Plugin Scheduler
        A Zookeeper Client
    """

    def __init__(self):
        super(PanoptesDiscoveryPluginSchedulerContext, self).__init__(
                key_value_store_class_list=[PanoptesDiscoveryPluginSchedulerKeyValueStore],
                create_message_producer=False, create_zookeeper_client=True)


class PanoptesCeleryDiscoveryAgentConfig(PanoptesCeleryConfig):
    task_routes = {const.DISCOVERY_PLUGIN_AGENT_MODULE_NAME: {'queue': const.DISCOVERY_PLUGIN_AGENT_CELERY_APP_NAME}}

    def __init__(self):
        super(PanoptesCeleryDiscoveryAgentConfig, self).__init__(
                app_name=const.DISCOVERY_PLUGIN_SCHEDULER_CELERY_APP_NAME)


def discovery_plugin_scheduler_task(celery_beat_service):
    """
    This function is the workhorse of the Discovery Plugin Scheduler module. It detects changes in plugins and their
    configuration and updates the Celery Beat schedule accordingly.

    Args:
        celery_beat_service (celery.beat.Service): The Celery Beat Service object associated with this Plugin Scheduler

    Returns:
        None
    """

    start_time = time.time()

    try:
        plugin_manager = PanoptesPluginManager(plugin_type='discovery',
                                               plugin_class=PanoptesDiscoveryPlugin,
                                               plugin_info_class=PanoptesPluginInfo,
                                               panoptes_context=panoptes_context,
                                               kv_store_class=PanoptesDiscoveryPluginSchedulerKeyValueStore)
        plugins = plugin_manager.getPluginsOfCategory(category_name='discovery')
    except:
        logger.exception('Error trying to load Discovery plugins, skipping cycle')
        return

    new_schedule = dict()

    for plugin in plugins:
        logger.info('Found plugin "%s", version %s at %s ' % (plugin.name, plugin.version, plugin.path))

        try:
            logger.info('Plugin "%s" has configuration: %s' % (plugin.name, plugin.config))
            logger.info('Plugin %s has plugin module time %s (UTC) and config mtime %s (UTC)' % (
                plugin.name, plugin.moduleMtime, plugin.configMtime))

            if plugin.execute_frequency <= 0:
                logger.info('Plugin %s has an invalid execution frequency (%d), skipping plugin' % (
                    plugin.name, plugin.execute_frequency))
                continue
        except PanoptesPluginConfigurationError as e:
            logger.error('Error reading/parsing configuration for plugin %s, skipping plugin. Error: %s' %
                         (plugin.name, repr(e)))

        logger.debug('Going to add task for plugin "%s" with execute frequency %d and args "%s"' % (
            plugin.name, plugin.execute_frequency, plugin.config))
        task_name = ':'.join([plugin.normalized_name, plugin.signature])
        new_schedule[task_name] = {
            'task': const.DISCOVERY_PLUGIN_AGENT_MODULE_NAME,
            'schedule': timedelta(seconds=plugin.execute_frequency),
            'args': (plugin.name,),
            'options': {
                'expires': expires(plugin),
                'time_limit': time_limit(plugin)
            }
        }

    end_time = time.time()

    try:
        scheduler = celery_beat_service.scheduler
        scheduler.update(logger, new_schedule)
        logger.info('Scheduled %d tasks in %.2fs' % (len(new_schedule), end_time - start_time))
    except:
        logger.exception('Error in updating schedule for Polling Plugins')


def start_discovery_plugin_scheduler():
    """
    The entry point for the Discovery Plugin Scheduler

    This method creates a Panoptes Context and the Celery Instance for the Discovery Plugin Scheduler

    Returns:
        None
    """
    global discovery_plugin_scheduler, celery, logger, panoptes_context

    try:
        panoptes_context = PanoptesDiscoveryPluginSchedulerContext()
    except Exception as e:
        sys.exit('Could not create a Panoptes Context: %s' % (repr(e)))

    try:
        celery_config = PanoptesCeleryDiscoveryAgentConfig()
    except Exception as e:
        sys.exit('Could not create a Celery Config object: %s' % repr(e))

    discovery_plugin_scheduler = PanoptesPluginScheduler(panoptes_context, 'discovery', 'Discovery',
                                                         celery_config,
                                                         const.DISCOVERY_PLUGIN_SCHEDULER_LOCK_ACQUIRE_TIMEOUT,
                                                         discovery_plugin_scheduler_task)
    logger = panoptes_context.logger
    celery = discovery_plugin_scheduler.start()

    if not celery:
        sys.exit('Could not start Celery Beat Service')


@beat_init.connect
def celery_beat_service_started(sender=None, args=None, **kwargs):
    """
    This method is called after Celery Beat instantiates it's service

    Args:
        sender (celery.beat.Service): The Celery Beat Service which was started by Celery Beat
        args (dict): Arguments
        **kwargs (dict): Keyword Arguments

    Returns:
        None
    """
    sender.scheduler.panoptes_context = panoptes_context
    sender.scheduler.task_prefix = const.DISCOVERY_PLUGIN_SCHEDULER_CELERY_TASK_PREFIX
    discovery_plugin_scheduler.run(sender, args, **kwargs)


# This wrapper is to ensure that the Discovery Plugin Scheduler only executes when called from Celery - prevents against
# execution when imported from other modules (like Sphinx) or called from the command line

if get_calling_module_name() == const.CELERY_LOADER_MODULE:
    faulthandler.enable()
    start_discovery_plugin_scheduler()
